package felix.dstruct;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;

import felix.dstruct.StatOperator.OPType;


import tuffy.db.RDB;
import tuffy.infer.ds.GClause;
import tuffy.mln.Clause;
import tuffy.mln.Literal;
import tuffy.mln.Predicate;
import tuffy.mln.Term;
import tuffy.mln.Type;
import tuffy.mln.Clause.ClauseInstance;
import tuffy.ra.ConjunctiveQuery;
import tuffy.ra.Expression;
import tuffy.util.Config;
import tuffy.util.ExceptionMan;


/**
 * The clause object used in Felix, which extends
 * the Clause class in Tuffy to contain Felix-related
 * fields and methods.
 * 
 * @author Ce Zhang
 *
 */
public class FelixClause extends Clause{
	
	
	@SuppressWarnings("unchecked")
	public FelixClause clone(){
		FelixClause ret = new FelixClause();
		
		ret.bilits = (ArrayList<Literal>) this.bilits.clone();
		
		ret.constraints = new ArrayList<Expression>();
		for(Expression sub : this.constraints){
			ret.constraints.add(sub.clone());
		}
		
		ret.cost = this.cost;
		ret.existentialVars = (ArrayList<String>) this.existentialVars.clone();
		ret.exprWeight = this.exprWeight;
		ret.id = this.id;
		ret.instances = (ArrayList<ClauseInstance>) this.instances.clone();
		ret.isTemplate = this.isTemplate;
		ret.lits = (ArrayList<Literal>) this.lits.clone();
		ret.metaTypes = (ArrayList<Type>) this.metaTypes.clone();
		ret.metaVars = (ArrayList<String>) this.metaVars.clone();
		ret.name = this.name;
		ret.predIndex = (HashMap<Predicate, ArrayList<Literal>>) this.predIndex.clone();
		ret.reglits = (ArrayList<Literal>) this.reglits.clone();
		ret.relIntanceClauses = this.relIntanceClauses;
		ret.signature = this.signature;
		ret.specText = (ArrayList<String>) this.specText.clone();
		ret.sqlFromList = this.sqlFromList;
		ret.sqlPivotAttrsList = this.sqlPivotAttrsList;
		ret.sqlWhereBindings = this.sqlWhereBindings;
		ret.uNames = (ArrayList<String>) this.uNames.clone();
		ret.varWeight = this.varWeight;
		ret.violatedGClauses = (ArrayList<GClause>) this.violatedGClauses.clone();
		ret.weight = this.weight;
		
		
		return ret;
	}
	
	/**
	 * @deprecated
	 */
	public boolean isBinaryLRRules = false;


	/**
	 * See {@link StatOperator#translateFelixClasesIntoFactorGraphEdgeQueries(FelixPredicate, boolean, dstruct.FelixPredicate.FPProperty...)}
	 * @param sop
	 * @param toBeHead
	 * @param forceRecursive
	 * @return
	 */
	public ConjunctiveQuery toSimplifiedFactorGraphQuery(StatOperator sop, FelixPredicate toBeHead, boolean forceRecursive){
		
		try{
			
			ConjunctiveQuery cq = new ConjunctiveQuery();
			
			//pick head literals
			int nOfTargetLiteral = 0;
			ArrayList<Term> headTerms = new ArrayList<Term>();
			ArrayList<Type> headTypes = new ArrayList<Type>();
			ArrayList<Literal> headLiterals = new ArrayList<Literal>();
			for(Literal l : this.getRegLiterals()){
				if(nOfTargetLiteral == 1 && forceRecursive == true){
					break;
				}
				
				if(l.getPred().getName().equals(toBeHead.getName())){
					nOfTargetLiteral ++;
					headTerms.addAll(l.getTerms());
					headLiterals.add(l);
					for(int i=0;i<l.getPred().arity();i++){
						headTypes.add(l.getPred().getTypeAt(i));
					}
				}
			}
			
			// currently, we do not consider hyper-graph edges
			if(nOfTargetLiteral > 2 && forceRecursive == false){
				ExceptionMan.die("Too many matched literals for predicate " + toBeHead);
			}
			
			FelixPredicate tmpHeadPredicate = new FelixPredicate(FelixPredicate.getNextTmpPredicateName(), false);
			for(Type type : headTypes){
				tmpHeadPredicate.appendArgument(type);
			}
			RDB db = RDB.getRDBbyConfig(Config.db_schema);
			tmpHeadPredicate.prepareDB(db);
			db.close();
			
			Literal lhead = new Literal(tmpHeadPredicate, true);
			for(Term t : headTerms){
				lhead.appendTerm(t);
			}
			cq.setHead(lhead);
			cq.sourceClause = this;
			
			int ctOpen = 0;
			boolean lastSense = false;
			Literal lastLiteral = null;
			for(Literal l : this.getRegLiterals()){
				if(headLiterals.contains(l)){
					continue;
				}
				ctOpen ++;
				Literal negL = (Literal) l.clone();
				negL.setSense(!negL.getSense());
				lastLiteral = negL;
				lastSense = negL.getSense();
				cq.addBodyLit(negL);
			}
			
			boolean flipFlag = false;
			if(ctOpen == 1 && lastSense == false){
				flipFlag = true;
				lastLiteral.setSense(!lastLiteral.getSense());
			}
			
			cq.setWeight(0);
			
			// set weight for the factor graph's edges
			if(headLiterals.size() == 1){
				if(headLiterals.get(0).getSense() == true){
					if(this.hasEmbeddedWeight()){
						cq.inverseEmbededWeight = (flipFlag == false? false : true);
					}else{
						cq.setWeight(flipFlag == false? this.getWeight(): -this.getWeight());
					}
				}else{
					if(this.hasEmbeddedWeight()){
						cq.inverseEmbededWeight = (flipFlag == false? true : false);
					}else{
						cq.setWeight((flipFlag == false?-this.getWeight():this.getWeight()));
					}					
				}
			}else{
				if(headLiterals.get(0).getSense() != headLiterals.get(1).getSense()){
					if(this.hasEmbeddedWeight()){
						cq.inverseEmbededWeight = (flipFlag == false? false : true);
					}else{
						cq.setWeight(flipFlag == false? this.getWeight(): -this.getWeight());
					}
				}else{
					if(this.hasEmbeddedWeight()){
						cq.inverseEmbededWeight = (flipFlag == false? true : false);
					}else{
						cq.setWeight((flipFlag == false?-this.getWeight():this.getWeight()));
					}
				}
			}
			
			if(sop.clauseConstraints.get(this) != null){
				cq.addConstraintAll(sop.clauseConstraints.get(this));
			}
			cq.addConstraintAll(this.getConstraints());
			
			return cq;
			
		}catch(Exception e){
			e.printStackTrace();
			return null;
		}

	}
	
}

