package felix.io;

import java.util.Arrays;

import org.python.core.PyCode;
import org.python.core.PyInteger;
import org.python.core.PyObject;
import org.python.util.PythonInterpreter;

/**
 * Class for executing python code (using Jython). 
 * @author Ce Zhang
 *
 */
public class PythonExecutor {

	/**
	 * pycode
	 */
	PyCode pycode;
	
	/**
	 * python interpreter
	 */
	PythonInterpreter interp;
	
	/**
	 * Prepares given script for execution.
	 * @param script
	 */
	public void prepare(String script){
		interp = new PythonInterpreter();
		pycode = interp.compile(script);
		
		interp.exec("def felixio_push(*columns):\n" + 
                    "\tif(len(columns) == 0):\n"+
                    "\t\t_felix_donotusemyname_outkey.append('')\n" + 
                    "\telse:\n" +
                    "\t\t_felix_donotusemyname_outkey.append(columns[0].replace('\\\\','\\\\\\\\').replace('\\t','\\\\t').replace('\\n', '\\\\n').replace('\\r', '\\\\r'))\n" +
                    "\t_felix_donotusemyname_outvalue.append('\t'.join(i.__str__().replace('\\\\','\\\\\\\\').replace('\\t','\\\\t').replace('\\n', '\\\\n').replace('\\r', '\\\\r') for i in columns[1:]))\n");
    
		interp.exec("def felixio_collect(k, v):\n" + 
                "\t_felix_donotusemyname_outkey.append(k)\n"+
                "\t_felix_donotusemyname_outvalue.append(v)");
		
	}
	
	/**
	 * Returns given variable.
	 * @param variable
	 * @return
	 */
	public PyObject get(String variable){
		return interp.get(variable);
	}
	
	/**
	 * Sets given variable.
	 * @param name
	 * @param value
	 */
	public void set(String name, PyObject value){
		interp.set(name, value);
	}
	
	/**
	 * Execuates given script.
	 * @param script
	 */
	public void execSingle(String script){
		interp.exec(script);
	}

	/**
	 * The constructor.
	 */
	public PythonExecutor(){
	}
	
	/**
	 * The constructor with script.
	 * @param script
	 */
	public PythonExecutor(String script){
		this.prepare(script);
	}
	
	/**
	 * Excecutes python script.
	 */
	public void run(){
		
		interp.exec(pycode);
		
	}
	
	/**
	 * Main test method.
	 * @param args
	 * @throws Exception
	 */
	/*
	public static void main(String[] args) throws Exception{
		
		PythonExecutor pe = new PythonExecutor();
		pe.prepare("print sys");
		
		pe.execSingle("import sys");
		
		for(int i=0;i<10000;i++){
			pe.run();
		}	
	}
	*/
	
}
