package felix.io;
import edu.umd.cloud9.collection.XMLInputFormat;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.StringTokenizer;
import org.apache.hadoop.streaming.StreamInputFormat;
import org.apache.hadoop.streaming.StreamXmlRecordReader;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.FileUtil;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.*;
import org.apache.hadoop.mapred.JobClient;
import org.apache.hadoop.mapred.OutputCollector;
import org.apache.hadoop.mapred.Reporter;
import org.apache.hadoop.mapred.lib.db.DBInputFormat;

import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.Reducer;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.input.FileSplit;
import org.apache.hadoop.mapreduce.lib.input.TextInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.apache.hadoop.mapreduce.lib.output.TextOutputFormat;


import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;
import org.python.core.PyArray;
import org.python.core.PyList;
import org.python.core.PyLong;
import org.python.core.PyObject;
import org.python.core.PyString;

import tuffy.util.ExceptionMan;

import felix.thirdpart.XmlInputFormat;
import felix.util.FelixConfig;

/**
 * This class communicates with Hadoop. It translates
 * the given MAP and REDUCE codes to Hadoop API.
 * @author czhang
 *
 */
public class TestHadoop extends Configured implements Tool{

	/**
	 * MAP
	 * @author czhang
	 *
	 */
	public static class Map extends Mapper<LongWritable, Text, Text, IntWritable>{
		
		private Text word = new Text();
		public static PythonExecutor pyMap;
		
		public static String _inputvalue;
		
		@Override
		public void setup(Mapper.Context contex) {

			pyMap = new PythonExecutor(
					"_felix_donotusemyname_outkey=[]\n" + "_felix_donotusemyname_outvalue=[]\n"+
					contex.getConfiguration().get("pyMapScript"));
			_inputvalue = contex.getConfiguration().get("mapinputvalue");
			
			if(contex.getConfiguration().get("pyMapInitScript") != null){
				pyMap.execSingle(contex.getConfiguration().get("pyMapInitScript"));
			}
			
		}

		public void map(LongWritable key, Text value, Mapper.Context context) throws IOException, InterruptedException {	
			
			FileSplit fileSplit = (FileSplit)context.getInputSplit();
			String fileName = fileSplit.getPath().getName();
			
			pyMap.set("_offset", new PyLong(key.toString()));
			pyMap.set("_filename", new PyString(fileName.toString()));			
			pyMap.set(_inputvalue, new PyString(value.toString()));
			
			pyMap.run();
			PyList outKey = (PyList) pyMap.get("_felix_donotusemyname_outkey");
			PyList outValue = (PyList) pyMap.get("_felix_donotusemyname_outvalue");
			
			for(int i=0;i<outKey.__len__();i++){
				PyObject _key = outKey.__getitem__(i);
				PyObject _value = outValue.__getitem__(i);
				
				context.write(new Text(""+_key), new Text(""+_value));
			//	System.out.println(""+key + "\t" + value);
			}
			
		}
		
	}
	
	/**
	 * REDUCE
	 * @author czhang
	 *
	 */
	public static class Reduce extends Reducer<Text, Text, Text, Text>{
		
		public static PythonExecutor pyReducer;
		
		public static String _inputkey;
		public static String _inputvalues;
		
		@Override
		public void setup(Reducer.Context contex) {
			pyReducer = new PythonExecutor(
					"_felix_donotusemyname_outkey=[]\n" + "_felix_donotusemyname_outvalue=[]\n" +
					contex.getConfiguration().get("pyReduceScript"));
			_inputkey = contex.getConfiguration().get("reduceinputkey");
			_inputvalues = contex.getConfiguration().get("reduceinputvalues");
			
			if(contex.getConfiguration().get("pyReduceInitScript") != null){
				pyReducer.execSingle(contex.getConfiguration().get("pyReduceInitScript"));
			}
			
		}

		protected void reduce(Text _key, Iterable _values, Context context) throws IOException, InterruptedException{
			
		
			ArrayList<PyString> args = new ArrayList<PyString>();
			Iterator it = _values.iterator();
			while(it.hasNext()){
				args.add(new PyString(it.next().toString()));
			}
						
			PyList values = new PyList(args);
			PyObject inputkey  = new PyString(_key.toString());
			
			pyReducer.set(_inputvalues, values);
			pyReducer.set(_inputkey, inputkey);
			
			pyReducer.run();
			PyList outKey = (PyList) pyReducer.get("_felix_donotusemyname_outkey");
			PyList outValue = (PyList) pyReducer.get("_felix_donotusemyname_outvalue");
			
			for(int i=0;i<outKey.__len__();i++){
				PyObject key = outKey.__getitem__(i);
				PyObject value = outValue.__getitem__(i);
				context.write(new Text("True\t1\t"+key), new Text(""+value));
			}
			
		}
		
	}
	
	/**
	 * @deprecated
	 * @param mapScript
	 * @param reduceScript
	 */
	public TestHadoop(String mapScript, String reduceScript){
		Map.pyMap = new PythonExecutor(
				mapScript);
		
		Reduce.pyReducer = new PythonExecutor(
				reduceScript);
	}
	
	/**
	 * Execute a given task configuration (via arg0).
	 * @param toPass
	 * @throws Exception
	 */
	public static void executeHadoopProgram(String[] toPass) throws Exception{
		
		String[] dirtoPass = toPass.clone();
		dirtoPass[2] += "_dir";
		int res = ToolRunner.run(new Configuration(), new TestHadoop(dirtoPass[3], dirtoPass[4]), 
				dirtoPass);
				
	}
	
	/**
	 * @deprecated
	 * @param toPass
	 * @throws Exception
	 */
	public static void post(String[] toPass) throws Exception{
		
		Configuration fsconf = new Configuration();
		fsconf.set("fs.default.name", FelixConfig.hdfsServer);
        FileSystem fileSystem = FileSystem.get(fsconf);
		
		String[] dirtoPass = toPass.clone();
		dirtoPass[2] += "_dir";
		
		FileUtil.copyMerge(fileSystem, new Path(dirtoPass[2]), fileSystem, new Path(toPass[2]), true,
				fsconf, "");
		
		fileSystem.close();
			
	}
	
	
	/**
	 * Test entry.
	 * @param args
	 * @throws Exception
	 */
	public static void main(String[] args) throws Exception{
	
		String[] toPass = {
				
				"hdfs://d-02.cs.wisc.edu:9000/felixNE/Entity.db",
				//"hdfs://d-02.cs.wisc.edu:9000/firstMapReduceOut42/part-00000",
				
				"hdfs://d-02.cs.wisc.edu:9000/firstMapReduceOut100",
				
				"for k in _input.split(' '):\n"+
				"\t_outkey.append(k)\n"+
				"\t_outvalue.append('1')\n",
				
				"_outkey.append(_inputkey)\n"+
				"_outvalue.append(len(_inputvalues))\n"
				
		};
		
		TestHadoop.executeHadoopProgram(toPass);
		
	
	}


	@Override
	/**
	 * Execute a given task configuration (via arg0).
	 */
	public int run(String[] arg0) throws Exception {
				

		Configuration conf = getConf();
		
		if(!FelixConfig.hadoopLocal){
			conf.set("fs.defaultFS", FelixConfig.hdfsServer);
			conf.set("mapred.job.tracker", FelixConfig.mrServer);
			conf.set("mapred.child.java.opts", "-Xmx3192m");
		}
		conf.set("pyMapScript", arg0[3]);
		conf.set("pyReduceScript", arg0[4]);
		conf.set("mapinputvalue", arg0[5]);
		conf.set("reduceinputkey", arg0[6]);
		conf.set("reduceinputvalues", arg0[7]);
		conf.set("pyMapInitScript", arg0[8]);
		conf.set("pyReduceInitScript", arg0[9]);
	
		
		if(arg0[0].equals("xml")){
			conf.set("xmlinput.start", arg0[10]);
			conf.set("xmlinput.end", arg0[11]);
		}
		
		Job job = new Job(conf, "Felix_Run_On" + (new Date()).toLocaleString());

		job.setNumReduceTasks(FelixConfig.nReduce);
		
		job.setJarByClass(TestHadoop.class);
		
		job.setMapperClass(Map.class);
		job.setReducerClass(Reduce.class);

		job.setOutputKeyClass(Text.class);
		job.setOutputValueClass(Text.class);

		FileInputFormat.addInputPath(job, new Path(arg0[1]));
		FileOutputFormat.setOutputPath(job, new Path(arg0[2]));
		
		
		if(arg0[0].equals("xml")){
			//XML
						
			job.setInputFormatClass(XmlInputFormat.class);
			job.setOutputFormatClass(TextOutputFormat.class);
			
			job.setOutputKeyClass(Text.class);
			job.setOutputValueClass(Text.class);
			
		}else if(arg0[0].equals("standard")){
			
			job.setInputFormatClass(TextInputFormat.class);
			job.setOutputFormatClass(TextOutputFormat.class);
			
			job.setOutputKeyClass(Text.class);
			job.setOutputValueClass(Text.class);
		}
		
		job.waitForCompletion(true);
		return 0;
	}
	
}
