package felix.main;


import java.util.HashMap;
import java.util.HashSet;



import tuffy.db.RDB;
import tuffy.mln.Clause;
import tuffy.mln.Predicate;
import tuffy.ra.ConjunctiveQuery;
import tuffy.util.ExceptionMan;
import tuffy.util.Timer;
import felix.compiler.StaticAnalyzer;
import felix.dstruct.ExecutionPlan;
import felix.dstruct.FelixQuery;
import felix.dstruct.OperatorBucketGraph;
import felix.dstruct.StatOperator;
import felix.executor.Executor;
import felix.optimizer.Scheduler;
import felix.parser.FelixCommandOptions;
import felix.util.FelixConfig;
import felix.util.FelixUIMan;

/**
 * 
 * 
 * @deprecated
 * @author Ce Zhang
 *
 */
public class OnlineFelix {

	/**
	 * Felix's query, which consists of program, query and evidence.
	 */
	FelixQuery fq;
	
	/**
	 * Static Analyzer, which analyzes the property of each predicates according to rules.
	 */
	StaticAnalyzer sa;
	
	/**
	 * Database connection.
	 */
	RDB db;
	
	/**
	 * Command line option.
	 */
	FelixCommandOptions options;
	
	/**
	 * Whether this Felix object run has loaded evidences.
	 */
	public boolean hasLoadedEvidence = false;
	
	/**
	 * Return Felix's query.
	 * @return
	 */
	public FelixQuery getFelixQuery(){
		return fq;
	}
	
	/**
	 * Clear static parameters and initialize variables.
	 */
	public void init(){
		
		ConjunctiveQuery.clearIndexHistory();
		
		Clause.mappingFromID2Const = new HashMap<Integer, String>();
		Clause.mappingFromID2Desc = new HashMap<String, String>();

	}
	
	/**
	 * Load Felix query from program and query file. If -useEvid
	 * is used, evidence file will also be loaded by this function.
	 */
	public void parseFelixQuery(){

		FelixUIMan.println(">>> Connecting to RDBMS at " + FelixConfig.db_url);
		
		fq = new FelixQuery();
		
		if(options.fquery == null && options.queryAtoms == null){
			System.err.println("Please specify queries with -q or -queryFiles");
			return;
		}
		
		FelixConfig.verbose_level = options.verboseLevel;

		String prog = options.fprog;
		fq.parseProgFromString(prog);

		if(options.fquery != null){
			String[] queryFiles = options.fquery.split(",");
			fq.loadQueries(queryFiles);
		}
		
		if(options.queryAtoms != null){
			FelixUIMan.println(">>> Parsing query atoms in command line");
			fq.parseQueryCommaList(options.queryAtoms);
		}

		if(options.cwaPreds != null){
			String[] preds = options.cwaPreds.split(",");
			for(String ps : preds){
				Predicate p = fq.getPredByName(ps);
				if(p == null){
					fq.closeFiles();
					ExceptionMan.die("COMMAND LINE: Unknown predicate name -- " + ps);
				}else{
					p.setClosedWorld(true);
				}
			}
		}
		
	}
	
	/**
	 * Run Felix!
	 * @param args Command line options.
	 */
	public String run(String[] args){
		
		FelixConfig.overrideID();
		String ret = "";
		
		Timer.start("Felix-Timer");
		
		init();
		
		options = FelixUIMan.parseCommand(args);

		this.parseFelixQuery();
		
		sa = new StaticAnalyzer(this.fq, options);
		sa.parse();
			
		//for(int j=0;j<2;j++){
		//
		Scheduler sc = new Scheduler(null, this.fq, options);
		
		// decompose the whole MLN into different operators
		HashSet<StatOperator> ops = sc.ruleDecomposition(null);

		// decompose each operators into smaller operators dealing
		// with different portion of data.
		OperatorBucketGraph obg = sc.dataDecomposition(ops, null);
		
		ExecutionPlan ep = sc.orderOperators(obg);

		ret = ep.toHTMLString();
		
		return ret;
		
	}
	
}




