package felix.operator;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.sql.Array;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;

import org.apache.commons.collections.Bag;
import org.apache.commons.collections.bag.HashBag;
import org.postgresql.PGConnection;


import tuffy.db.RDB;
import tuffy.mln.Literal;
import tuffy.ra.ConjunctiveQuery;
import tuffy.util.Config;
import tuffy.util.ExceptionMan;
import tuffy.util.FileMan;
import tuffy.util.StringMan;
import tuffy.util.Timer;
import tuffy.util.UIMan;
import felix.dstruct.DataMovementOperator;
import felix.dstruct.FelixPredicate;
import felix.dstruct.FelixQuery;
import felix.dstruct.StatOperator;
import felix.dstruct.FelixPredicate.FPProperty;
import felix.parser.FelixCommandOptions;
import felix.util.FelixConfig;
import felix.util.FelixStringMan;
import felix.util.FelixUIMan;

/**
 * A LR operator in Felix.
 * @author Ce Zhang
 *
 */
public class LROperator extends StatOperator{
	
	/**
	 * Target predicate of this LR operator.
	 */
	FelixPredicate lrHead;
	

	/**
	 * Mapping from label's constant ID to a new label ID. This new label ID is from 0, which 
	 * is used in the inference of LR (s.t., we can use array to represent
	 * labels).
	 */
	HashMap<String, Integer> label2ID = new HashMap<String, Integer>();
	
	/**
	 * The inverse map of {@link LROperator#label2ID}.
	 */
	HashMap<Integer, String[]> id2Label = new HashMap<Integer, String[]>();
	
	/**
	 * All DataMovementOperators used as LR rules.
	 */
	ArrayList<DataMovementOperator> lrDMOs = new ArrayList<DataMovementOperator>();
	
	/**
	 * The DataMovementOperator which is the union of all LR rules.
	 */
	DataMovementOperator lrDMO = null;
	
	/**
	 * The DataMovementOperator representing the table/view for all unigram features.
	 */
	DataMovementOperator unigramDMO = null;
	
	/**
	 * The DataMovementOperator representing the table/view for all possible labels.
	 */
	DataMovementOperator labelDomainDMO = null;
	
	/**
	 * The DataMovementOperator fetching all unigram features.
	 */
	DataMovementOperator getAllUnigramFeaturesDMO = null;
	
	/**
	 * @deprecated
	 */
	DataMovementOperator getAllLearningFeatureDMO = null;
	
	/**
	 * @deprecated
	 */
	DataMovementOperator getGroundTruthDMO = null;
	
	/**
	 * @deprecated
	 */
	DataMovementOperator getFeatureTableDMO = null;
	
	/**
	 * The constructor of LROperator.
	 * @param _fq Felix query.
	 * @param _goalPredicates target predicates of this coref operator.
	 * @param _opt Command line options of this Felix run.
	 */
	public LROperator(FelixQuery _fq, HashSet<FelixPredicate> _goalPredicates,
			FelixCommandOptions _opt) {
		super(_fq, _goalPredicates, _opt);
		this.type = OPType.LR;
		this.precedence = 4;
		
	}
	
	/**
	 * Adds object to domain if necessary.
	 * @param obj
	 * @param map
	 * @return
	 */
	public <E> Integer getAndAddToDomain(E obj, HashMap<String, Integer> map){
		int rs;
		if(map.containsKey(obj.toString())){
			return map.get(obj.toString());
		}
		rs = map.size() + 1;
		map.put(obj.toString(), rs);
		return rs;
	}
	
	/**
	 * @deprecated
	 */
	@Override
	public void learn() {
		
	}
	
	
	boolean prepared = false;
	
	/**
	 * Prepares operator for execution.
	 */
	@Override
	public void prepare() {
		
		lrDMOs.clear();
		allDMOs.clear();
		
		//if(!prepared){
			
			db = RDB.getRDBbyConfig(Config.db_schema);
		
			lrHead = this.getTargetPredicateIfHasOnlyOne();
			
			HashSet<ConjunctiveQuery> lrQueries;
			
			if(options.isDLearningMode){
				//lrQueries = 
				//	this.translateFelixClasesIntoLearningQueriesForVictor(lrHead, FPProperty.NON_RECUR);
				//this.prepareDMO4ForLearning(lrQueries);
				
			}else{
				lrQueries = 
					this.translateFelixClasesIntoFactorGraphEdgeQueries(lrHead, false, this.inputPredicateScope, FPProperty.NON_RECUR);
				this.prepareDMO(lrQueries);
			}
		
			prepared = true;
			
			//db.close();
			
		//}
		
	}
	
	
	/**
	 * Returns string representation of string array. 
	 * @param _array
	 * @return
	 */
	String array2str(String[] _array){
		String ret = "";
		for(String s : _array){
			ret = ret + ":" + s;
		}
		return ret;
	}
	
	/**
	 * Executes operator.
	 */
	@Override
	public void run() {
		
		if(options.isDLearningMode){
			this.learn();
		}else{
		
			this.isMarginal = belongsToBucket.isMarginal();
			
			UIMan.println(">>> Start Running " + this);
			
			ArrayList<Integer> nonZeroWeights = new ArrayList<Integer>();
	
			try{
				
				Timer.start("LR-Operator-" + lrHead.getName());
				
				BufferedWriter bw = new BufferedWriter(
						new FileWriter(Config.getLoadingDir() + "/_loading_lr_" + lrHead.getName() + this.getId()));
						
				// get the domain of Label, and assign them an integer ID (start from 0, s.t., we
				// can use array to store them
				int labelID = 0;
				int nLabelFileds = lrHead.getLabelFieldsArgs().size();
				if(this.labelDomainDMO != null){
					this.labelDomainDMO.execute(null, new ArrayList<Integer>());
					while(this.labelDomainDMO.next()){
						String[] currLabel = new String[nLabelFileds];
						
						for(int i=0;i<nLabelFileds;i++){
							currLabel[i] = this.labelDomainDMO.getNext(i+1).toString();
						}
						
						this.id2Label.put(labelID, currLabel);
						this.label2ID.put(array2str(currLabel), labelID);
						labelID++;
					}
				}
				
				
				if(this.id2Label.size() == 0 && this.isBinaryArbLR == false){
					return;
				}
				
				if(this.unigramDMO == null){
					UIMan.warn("There are no LR-rules for predicate " + lrHead.getName());
					return;
				}
				
				int ct = 0;
				
				String workingSignature = null;
				this.getAllUnigramFeaturesDMO.db.disableAutoCommitForNow();
				this.getAllUnigramFeaturesDMO.execute(null, new ArrayList<Integer>());
				double[] weights;
				
				if(this.isBinaryArbLR == false){
					weights = new double[this.label2ID.keySet().size()];
				}else{
					weights = new double[1];
				}
					
				String[] currLabel = new String[lrHead.getLabelFieldsArgs().size()];
				
				while(this.getAllUnigramFeaturesDMO.next()){
							
					String currSignature = "";
					int lct = 0;
					
					ArrayList<String> toSig = new ArrayList<String>();
					if(this.isBinaryArbLR == false){
						for(int i=0;i<lrHead.arity();i++){
							if(lrHead.getLabelPositions().contains(i)){
								currLabel[lct++] = this.getAllUnigramFeaturesDMO.getNext(i+1).toString();
								toSig.add("%s");
							}else{
								toSig.add(this.getAllUnigramFeaturesDMO.getNext(i+1).toString());
								
								//if(this.getAllUnigramFeaturesDMO.getNext(i+1).toString().equals("264014")){
								//	System.out.println();
								//}
							}
						}
					}else{
						for(int i=0;i<lrHead.arity();i++){
							toSig.add(this.getAllUnigramFeaturesDMO.getNext(i+1).toString());
						}
					}
					
					currSignature = StringMan.commaList(toSig);
					Double weight = this.getAllUnigramFeaturesDMO.getNextDouble(lrHead.arity()+1);
					
					if(! currSignature.equals(workingSignature)){
						//dump and init. 
						if(workingSignature != null){
							if(ct % 10000 == 0){
								UIMan.verbose(3, "" + (ct++));
							}
							this.inferAndDumpToTmpFile(workingSignature, weights, nonZeroWeights, bw);
							for(int i : nonZeroWeights){
								weights[i] = 0;
							}
							nonZeroWeights.clear();
						}
						workingSignature = currSignature;
					}
					
					if(this.isBinaryArbLR == false){
						/*
						System.out.println("!  " + array2str(currLabel) + "");
						System.out.println("@  " + this.label2ID.get(array2str(currLabel)) + "");
						System.out.println("#  " + weight + "\n\n");
						*/
						currLabel = currLabel;
						array2str(currLabel);
						int wlen = this.label2ID.get(array2str(currLabel));
						//System.out.print(wlen + " | ");
						//System.out.print(weights.length + " | ");
						//System.out.println(weight);
						weights[this.label2ID.get(array2str(currLabel))] += weight;
						nonZeroWeights.add(this.label2ID.get(array2str(currLabel)));
					}else{
						weights[0] += weight;
					}
				}
				if(workingSignature != null){
					this.inferAndDumpToTmpFile(workingSignature, weights, nonZeroWeights, bw);
				}
				
				
				this.getAllUnigramFeaturesDMO.db.setAutoCommit(true);
				bw.close();
	
				//System.out.println("~\t" + Timer.elapsed("LR-Operator-" + lrHead.getName()));
				
				//UIMan.print("Start COPY...");
				FileInputStream in = new FileInputStream(Config.getLoadingDir() + "/_loading_lr_" + lrHead.getName() + this.getId());
				PGConnection con = (PGConnection)db.getConnection();
	
				String sql;
				
				if(options.useDualDecomposition){
					for(FelixPredicate fp : this.dd_CommonOutput){
						if(!fp.getName().equals(this.lrHead.getName())){
							ExceptionMan.die("LR 583: There must be something wrong with the parser!");
							continue;
						}
						
						in = new FileInputStream(Config.getLoadingDir() + 
								"/_loading_lr_" + lrHead.getName() + this.getId());
						String tableName = this.dd_commonOutputPredicate_2_tableName.get(fp);
						
						
						sql = "COPY " + tableName + "(truth, prior, club, " + StringMan.commaList(lrHead.getArgs())  + " ) FROM STDIN CSV";
						con.getCopyAPI().copyIn(sql, in);
						in.close();
						
					}
					
					if(FelixConfig.isFirstRunOfDD){
						in = new FileInputStream(Config.getLoadingDir() + 
								"/_loading_lr_" + lrHead.getName() + this.getId());
						
						sql = "COPY " + lrHead.getRelName() + "(truth, prior, club, " + StringMan.commaList(lrHead.getArgs())  + " ) FROM STDIN CSV";
						con.getCopyAPI().copyIn(sql, in);
						in.close();
					}
					lrHead.isCurrentlyView = false;
					
				}else{
					sql = "COPY " + lrHead.getRelName() + "(truth, prior, club, " + StringMan.commaList(lrHead.getArgs())  + " ) FROM STDIN CSV";
					con.getCopyAPI().copyIn(sql, in);
					in.close();
					lrHead.isCurrentlyView = false;
				}
	
				FelixUIMan.println(0,0,"\n>>> {" + this + "} uses " + Timer.elapsed("LR-Operator-" + lrHead.getName()));
				
			//	for(FelixPredicate fp : this.outputPredicates){
			//		fp.setClosedWorld(true);
			//	}
				
				lrHead.setHasSoftEvidence(true);
			
			
				this.db.close();
				
				if(!options.useDualDecomposition){
					this.belongsToBucket.runNextOperatorInBucket();
				}
				
				
				return;
				
			}catch(Exception e){
				e.printStackTrace();
			}
		}
		
	}

	@Override
	public String explain() {

		return null;
	}

	/**
	 * Generate Data Movement Operator used by this LR Operator.
	 * @param rules rules defining this operator.
	 */
	public void prepareDMO(HashSet<ConjunctiveQuery> lrQueries){
		
		try {
			
			// DMO for LR rules
			for(ConjunctiveQuery cq : lrQueries){
				
				DataMovementOperator dmo = new DataMovementOperator(db, this);
				dmo.logicQueryPlan.addQuery(cq, cq.head.getPred().getArgs(),
						new ArrayList<String>(Arrays.asList("weight", "prov", "deepprov")) );
				
				dmo.predictedBB = 0;
				dmo.PredictedFF = 1;
				dmo.PredictedBF = 0;
				
				dmo.allowOptimization = false;
				
				allDMOs.add(dmo);
				lrDMOs.add(dmo);
			}
			
			//the DMO for the union of all LR DMOs
			if(this.lrDMOs.size() > 0){
				this.lrDMO = DataMovementOperator.UnionAll(db, this, 
						this.lrDMOs, StringMan.zeros(this.lrHead.arity() + 3), new ArrayList<Integer>());
				allDMOs.add(lrDMO);
				
		//		if(!db.isTableExists(FelixConfig.db_schema, "_prov_metainfo")){
		//			db.execute("CREATE TABLE _prov_metainfo (predicate TEXT, property TEXT, value TEXT)");
		//		}
			//	
			//	db.execute("INSERT INTO _prov_metainfo (predicate, property, value) VALUES ('" + lrHead.getName() 
			//			+ "', 'optype', '" + "lr" + "')");
			//	db.execute("INSERT INTO _prov_metainfo (predicate, property, value) VALUES ('" + lrHead.getName() 
			//			+ "', 'provtable', '" + lrDMO.getAllFreeViewName()+ "')");
			//	db.execute("INSERT INTO _prov_metainfo (predicate, property, value) VALUES ('" + lrHead.getName() 
			//			+ "', 'labelfield', '" + StringMan.commaList(lrHead.getLabelFieldsArgs())+ "')");
				
								
				//the DMO for all unigram features.
				this.unigramDMO = new DataMovementOperator(db, this);
				this.unigramDMO.allowOptimization = false;
				this.unigramDMO.asView = false;
				this.unigramDMO.logicQueryPlan.addQuery(db.getPrepareStatement(
						"SELECT " + StringMan.commaList(lrDMO.selListFromRule) + ", sum(weight) AS sumweight " +
								" FROM " + lrDMO.getAllFreeViewName() + " GROUP BY " + StringMan.commaList(lrDMO.selListFromRule) + 
								(lrHead.getKeyFieldsArgs().size() == 0 ? "" :
										" ORDER BY " + StringMan.commaList(lrHead.getKeyFieldsArgs())) ),
								lrDMO.selListFromRule, new ArrayList<String>(Arrays.asList("sumweight")));
				allDMOs.add(this.unigramDMO);
			}		
			
			//the DMO for the label domain
			if(this.isBinaryArbLR == false){
				this.labelDomainDMO = new DataMovementOperator(db, this);
				this.labelDomainDMO.allowOptimization = false;
				ArrayList<String> fields = lrHead.getLabelFieldsTypeTable();
				ArrayList<String> fieldsViewName = new ArrayList<String>();
				for(int i=0;i<fields.size();i++){
					fieldsViewName.add("t" + i + "." + "constantid AS l" + i);
					fields.set(i, fields.get(i) + " t" + i);
				}
				this.labelDomainDMO.logicQueryPlan.addQuery(db.getPrepareStatement(
						"SELECT " + StringMan.commaList(fieldsViewName) +" FROM " + StringMan.commaList(fields)),
								lrHead.getLabelFieldsArgs(), 
								new ArrayList<String>());
				allDMOs.add(this.labelDomainDMO);
			}
				
			if(this.unigramDMO != null){
				this.getAllUnigramFeaturesDMO = DataMovementOperator.Select(db, this, 
						this.unigramDMO, new ArrayList<String>());
				this.getAllUnigramFeaturesDMO.isIntermediaDMO = true;
				this.getAllUnigramFeaturesDMO.hasKnownFetchingOrder = true;
				allDMOs.add(this.getAllUnigramFeaturesDMO);
			}
		
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Returns sum of given log numbers.
	 * @param logX
	 * @param logY
	 * @return
	 */
	//ACKNOWLEDGE: FROM https://facwiki.cs.byu.edu/nlp/index.php/Log_Domain_Computations
	//COPYRIGHT OF THIS FUNCTION BELONGS TO ITS ORIGINAL AUTHOR
	public static double logAdd(double logX, double logY) {

	       if (logY > logX) {
	           double temp = logX;
	           logX = logY;
	           logY = temp;
	       }

	       if (logX == Double.NEGATIVE_INFINITY) {
	           return logX;
	       }
	       
	       double negDiff = logY - logX;
	       if (negDiff < -200) {
	           return logX;
	       }
	       
	       return logX + java.lang.Math.log(1.0 + java.lang.Math.exp(negDiff)); 
	 }
	
	/**
	 * Infer and dump answers to the given buffered writer (with a format that can be COPY
	 * into postgres table directly).
	 * @param signature signature of current predicate instance.
	 * @param weights array of weights. Each entry corresponding to the weight of a label.
	 * @param bw
	 */
	public void inferAndDumpToTmpFile(String signature, double[] weights, ArrayList<Integer> noneZeroWeights, BufferedWriter bw){
		
		try{

			double sum = Double.NEGATIVE_INFINITY;
			double tmp = Double.NEGATIVE_INFINITY;
			double max = Double.NEGATIVE_INFINITY;
			int nmax = -1;
			
			for(int i : noneZeroWeights){		
//			for(int i=0;i<weights.length;i++){	
				tmp = weights[i];
				if(tmp > max){
					max = tmp;
					nmax = i;
				}
				sum = logAdd(tmp, sum);
			}
			
			if(max < 0){
				for(int i=0;i<weights.length;i++){		
						if(weights[i] >= 0){
							max = weights[i];
							nmax = i;
							break;
						}
					}
			}
			
			sum = logAdd(sum, Math.log(this.label2ID.size() - noneZeroWeights.size()));
			
			if(this.isBinaryArbLR == true){
				sum = logAdd(0,sum);
			}
			
			if(this.isMarginal || (FelixConfig.isFirstRunOfDD && dd_commonOutputPredicate_2_tableName.containsKey(lrHead) )  ){
				
				if(nmax == -1){
					return;
				}
				
				for(int i : noneZeroWeights){	
//				for(int i=0;i<weights.length;i++){
				
					tmp = weights[i];
					
					double prob = Math.exp(tmp - sum);
					
					if(prob >= Config.soft_evidence_activation_threshold){
					
						ArrayList<String> parts = new ArrayList<String>();
						//parts.add(Integer.toString(lrHead.nextTupleIDAndUpdate()));
						parts.add("TRUE");
						parts.add(Double.toString(prob));
						
						if(options.useDualDecomposition){
							parts.add(Integer.toString(2));
						}else{
							parts.add(Integer.toString(2));
						}
						//parts.add("1");//this is for vote
						
						String tmpstr = String.format(signature, (Object[]) id2Label.get(i));
						tmpstr = tmpstr.replaceAll("INTERNAL_MARGINAL", prob+"");
						
						bw.append(FelixStringMan.commaListNoSpace(parts) + "," + tmpstr + "\n");
					}
				}
				
			}else{
				
				if(nmax == -1){
					return;
				}
				
//				double prob = Math.exp(weights[nmax]- sum);
	
				if(this.isBinaryArbLR == true){
					if(Math.exp(weights[0] - sum) <= 0.5){
						return;
					}
				}
				
				ArrayList<String> parts = new ArrayList<String>();
				//parts.add(Integer.toString(lrHead.nextTupleIDAndUpdate()));
				parts.add("TRUE");
				parts.add("");
				
				if(options.useDualDecomposition){
					parts.add(Integer.toString(2));
				}else{
					parts.add(Integer.toString(2));
				}
				
				
				//parts.add("1");//this is for vote
				
				String tmpstr = String.format(signature, (Object[]) id2Label.get(nmax));
				
				tmpstr = tmpstr.replaceAll("INTERNAL_MARGINAL", "1");
				
				bw.append(FelixStringMan.commaListNoSpace(parts) + "," + tmpstr + "\n");
				
			}
			
		}catch(Exception e){
			e.printStackTrace();
		}
		
	}




	
}
