package felix.operator;


import felix.dstruct.FelixClause;
import felix.dstruct.FelixPredicate;
import felix.dstruct.FelixQuery;
import felix.dstruct.StatOperator;
import felix.dstruct.FelixPredicate.FPProperty;
import felix.parser.FelixCommandOptions;
import felix.util.FelixConfig;
import felix.util.FelixUIMan;


import java.util.HashMap;
import java.util.HashSet;



import tuffy.db.RDB;
import tuffy.ground.Grounding;
import tuffy.ground.KBMC;
import tuffy.ground.partition.PartitionScheme;
import tuffy.infer.DataMover;
import tuffy.infer.InferPartitioned;
import tuffy.infer.MRF;
import tuffy.mln.Clause;
import tuffy.mln.MarkovLogicNetwork;
import tuffy.mln.Predicate;
import tuffy.ra.ConjunctiveQuery;
import tuffy.ra.Expression;
import tuffy.util.Config;
import tuffy.util.Settings;
import tuffy.util.Timer;
import tuffy.util.UIMan;


/**
 * A Tuffy operator in Felix.
 * @author Ce Zhang
 *
 */
public class TUFFYOperator extends StatOperator{
	
	/**
	 * The grounding worker of this Tuffy operator.
	 */
	Grounding grounding;
	
	/**
	 * Markov logic network used by this Tuffy operator.
	 */
	MarkovLogicNetwork mln;

	/**
	 * The constructor of TUFFYOperator.
	 * @param _fq Felix query.
	 * @param _goalPredicates target predicates of this coref operator.
	 * @param _opt Command line options of this Felix run.
	 */
	public TUFFYOperator(FelixQuery _fq, HashSet<FelixPredicate> _goalPredicates,
			FelixCommandOptions _opt) {
		super(_fq, _goalPredicates, _opt);
		this.type = OPType.TUFFY;
		this.precedence = 1;
		
		for(FelixClause fc : this.allRelevantFelixClause){
			
			int nOpen = 0;
					
			for(Predicate p : fc.getReferencedPredicates()){
				
				if(!p.isClosedWorld() || p.getName().endsWith("_map")){
					this.commonCandidate.add(fq.getPredByName(p.getName()));
				}
			}
				
		}
		
	}

	boolean prepared = false;
	
	/**
	 * Prepares operator for execution.
	 */
	@Override
	public void prepare() {
		
		if(!prepared){
			
			UIMan.println(">>> Start Running " + this);
			
			try {
				
				
			} catch (Exception e) {
				e.printStackTrace();
			}
			
			prepared = true;
		}
		
	}


	/**
	 * Executes operator.
	 */
	@Override
	public void run() {
		
		UIMan.println(">>> Start Running " + this);

		db = RDB.getRDBbyConfig(Config.db_schema);
		
		Timer.start("Timer - TUFFY - " + this.getId());
		
		mln = new MarkovLogicNetwork();
		
		mln.setDB(db);
		
		HashSet<Predicate> registeredPredicates = new HashSet<Predicate>();
		
		HashMap<Predicate, Boolean> oriClosedWorld = new HashMap<Predicate, Boolean>();
		

		for(FelixPredicate fp : this.inputPredicates){
			oriClosedWorld.put(fp, fp.isClosedWorld());
			fp.setClosedWorld(true);
			if(this.isMarginal){
				if(fp.isCorefPredicate == false && fp.isCorefMapPredicate == false){
					if(!options.useDualDecomposition || FelixConfig.isFirstRunOfDD){
						fp.setHasSoftEvidence(true);
					}else{
						fp.setHasSoftEvidence(false);
					}
				}
			}

		}
		
		for(Predicate fp : this.outputPredicates){
			oriClosedWorld.put(fp, fp.isClosedWorld());
			fp.setClosedWorld(this.currentState);
			fp.setCompeletelySpecified(this.currentState);

		}
		
		for(FelixPredicate fp : this.dd_CommonOutput){

			fp.setHasSoftEvidence(true);
			fp.setCompeletelySpecified(true);
			oriClosedWorld.put(fp, fp.isClosedWorld());
			fp.setClosedWorld(!this.currentState);

		}
		
		HashSet<FelixClause> allClause = allRelevantFelixClause;
		allClause.addAll(this.dd_PriorClauses);
		
		for(FelixClause fc : this.allRelevantFelixClause){
			
			boolean isNotUseful = true;
			
			Clause cloned = fc.clone();
			
			if(this.clauseConstraints.get(fc) != null){
			
				for(Expression e : this.clauseConstraints.get(fc)){
					cloned.getConstraints().add(e.clone());
				}
			}
			
			mln.registerClause(cloned);
			
			for(Predicate fp : fc.getReferencedPredicates()){
				if(registeredPredicates.contains(fp)){
					continue;
				}
				fp.getRelatedClauses().clear();
				mln.registerPred(fp);
				
				registeredPredicates.add(fp);
			}
			
		}
		
		for(ConjunctiveQuery sr : fq.getScopingRules()){
			mln.registerScopingRule(sr.clone());
		}
		
		mln.normalizeClauses();
		
		for(Predicate p : mln.getAllPred()){
			p.setDB(db);
		}
		mln.finalizeClauseDefinitions(db);
		
		
		//mln.prepareDB4Mobius(db);
		
		//mln.cleanUnknownPredTables();
		
		KBMC kbmc = new KBMC(mln);
		kbmc.run();
		mln.applyAllScopes();
		FelixUIMan.println(1, ">>> Marking queries...");
		mln.storeAllQueries();	
		
		mln.getDB().commit();
		//mln.getDB().close();
		//mln.setDB(RDB.getRDBbyConfig(Config.db_schema));
		
		grounding = new Grounding(mln);
		grounding.constructMRF();
		
		DataMover dmover = new DataMover(mln);
		
		if(options.maxFlips == 0){
			options.maxFlips = 10 * grounding.getNumAtoms();
		}
		if(options.maxTries == 0){
			options.maxTries = 1;
		}

		MRF mrf = null;
		
		if(options.disablePartition){
			if(!options.marginal || options.dual){
				UIMan.println(">>> Running MAP inference...");
				String mapfout = options.fout;
				if(options.dual) mapfout += ".map";
	
				UIMan.println("    Loading MRF from DB to RAM...");
				mrf = dmover.loadMrfFromDb(mln.relAtoms, mln.relClauses);
				mrf.inferWalkSAT(options.maxTries, options.maxFlips);
				dmover.flushAtomStates(mrf.atoms.values(), mln.relAtoms, true);
				
				UIMan.println("### Best answer has cost " + UIMan.decimalRound(2,mrf.lowCost));
				//UIMan.println(">>> Writing answer to file: " + mapfout);
				//dmover.dumpTruthToFile(mln.relAtoms, mapfout);
			}
	
			if(options.marginal || options.dual){
				UIMan.println(">>> Running marginal inference...");
				String mfout = options.fout;
				if(options.dual) mfout += ".marginal";
				
				if(mrf == null){
					mrf = new MRF(mln);
					dmover.loadMrfFromDb(mrf, mln.relAtoms, mln.relClauses);
				}
				
				double sumCost = mrf.mcsat(options.mcsatSamples, options.maxFlips);
				dmover.flushAtomStates(mrf.atoms.values(), mln.relAtoms);
				
				UIMan.println("### Average Cost = " + UIMan.decimalRound(2,sumCost/options.mcsatSamples));
				
				//UIMan.println(">>> Writing answer to file: " + mfout);
				//dmover.dumpProbsToFile(mln.relAtoms, mfout);
			}
		}else{
			
			InferPartitioned ip = new InferPartitioned(grounding, dmover);
			PartitionScheme pmap = ip.getPartitionScheme();
			int ncomp = pmap.numComponents();
			int nbuck = ip.getNumBuckets();
			String sdata = UIMan.comma(ncomp) + (ncomp > 1 ? " components" : "component");
			sdata += " (grouped into ";
			sdata += UIMan.comma(nbuck) + (nbuck > 1 ? " buckets" : " bucket)");
			
			
			Settings settings = new Settings(); 
			Double fpa = ((double)options.maxFlips)/grounding.getNumAtoms();
			
			if(!options.marginal || options.dual){
				UIMan.println(">>> Running MAP inference on " + sdata);
				String mapfout = options.fout;
				if(options.dual) mapfout += ".map";
				
				settings.put("task", "MAP");
				settings.put("ntries", new Integer(options.maxTries));
				settings.put("flipsPerAtom", fpa);
				double lowCost = ip.infer(settings);
				
				UIMan.println("### Best answer has cost " + UIMan.decimalRound(2,lowCost));
				//UIMan.println(">>> Writing answer to file: " + mapfout);
				//dmover.dumpTruthToFile(mln.relAtoms, mapfout);
			}
			
			if(options.marginal || options.dual){
				UIMan.println(">>> Running marginal inference on " + sdata);
				String mfout = options.fout;
				if(options.dual) mfout += ".marginal";
				
				settings.put("task", "MARGINAL");
				settings.put("nsamples", new Integer(options.mcsatSamples));
				settings.put("flipsPerAtom", fpa);
				double aveCost = ip.infer(settings);
				
				UIMan.println("### Average Cost = " + UIMan.decimalRound(2,aveCost));
				
				//UIMan.println(">>> Writing answer to file: " + mfout);
				//dmover.dumpProbsToFile(mln.relAtoms, mfout);
			}
			
		}

		this.belongsToBucket.addMLNRelTable(mln.relAtoms);
		
		for(Predicate fp : oriClosedWorld.keySet()){
			fp.setClosedWorld(oriClosedWorld.get(fp));
		}
		
		FelixUIMan.println(0,0,">>> {" + this + "} uses " + Timer.elapsed("Timer - TUFFY - " + this.getId()));
		FelixUIMan.println(0, 0, "");
		
		//System.out.println("--------------[Timer - TUFFY - " + this.getId() + "]: " + 
		//		Timer.elapsed("Timer - TUFFY - " + this.getId()));
				
		db.close();
		
		if(!options.useDualDecomposition){
			this.belongsToBucket.runNextOperatorInBucket();
		}
		
	}

	@Override
	public String explain() {
		return null;
	}

	@Override
	public void learn() {
		
	}

}
