package felix.optimizer;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;

import tuffy.mln.Literal;
import tuffy.mln.Predicate;
import tuffy.ra.Expression;
import tuffy.ra.Function;

import org.apache.commons.math.optimization.GoalType;
import org.apache.commons.math.optimization.RealPointValuePair;
import org.apache.commons.math.optimization.linear.LinearConstraint;
import org.apache.commons.math.optimization.linear.LinearObjectiveFunction;
import org.apache.commons.math.optimization.linear.Relationship;
import org.apache.commons.math.optimization.linear.SimplexSolver;

import felix.dstruct.FelixClause;
import felix.dstruct.FelixPredicate;
import felix.dstruct.StatOperator;

//TODO: CURRENTLY SOLVED BY ROUND!!!
//TODO: MAYBE CHANGE TO HASH!!!!!!


/**
 * An object of DataCracker1991 will try its best to discover opportunities
 * of partitioning the data.
 */
public class DataCracker1991 {

	/**
	 * Map from variable ID to the signature of predicate's field. This ID is used
	 * in the ILP solver.
	 */
	public HashMap<Integer, String> varID2predField = new HashMap<Integer, String>();
	
	/**
	 * Map from variable ID to the predicate. This ID is used
	 * in the ILP solver.
	 */
	public HashMap<Integer, FelixPredicate> varID2pred = new HashMap<Integer, FelixPredicate>();
	
	/**
	 * Map from variable ID to predicate's field ID. This ID is used
	 * in the ILP solver.
	 */
	public HashMap<Integer, Integer> varID2field = new HashMap<Integer, Integer>();
	
	/**
	 * The inverse map of {@link DataCracker1991#varID2predField}.
	 */
	public HashMap<String, Integer> predField2varID = new HashMap<String, Integer>();
	
	/**
	 * The array that saves the solution of the ILP solver.
	 */
	double[] ss;
	
	/**
	 * Whether the recently processed operator is decomposable.
	 */
	boolean isDecomposable = false;
	
	/**
	 * Generate the signature of predicate's field.
	 * @param p
	 * @param fieldNum
	 * @return
	 */
	public String toCanonicalFieldName(FelixPredicate p, int fieldNum){
		return p.getName() + ":" + p.getArgs().get(fieldNum);
	}
	
	/**
	 * Generate the signature of predicate's field.
	 * @param p
	 * @param fieldNum
	 * @return
	 */
	public String toCanonicalFieldName(Predicate p, int fieldNum){
		return p.getName() + ":" + p.getArgs().get(fieldNum);
	}
	
	/**
	 * Decompose the data used by the given operator.
	 * @param op
	 */
	public void decompose(StatOperator op){
		
		try{
			
			// get all open predicates can be seperated
			HashSet<FelixPredicate> openPredicates = op.outputPredicates;
			
			int ct = 0;
			for(FelixPredicate p : openPredicates){		
				for(int i=0;i<p.arity();i++){
					varID2predField.put(ct, toCanonicalFieldName(p, i));
					varID2pred.put(ct, p);
					varID2field.put(ct, i);
					predField2varID.put(toCanonicalFieldName(p, i), ct);
					ct++;
				}
			}
			
			
			double[] zeros = new double[ct];
			double[] ones = new double[ct];
			for(int i=0;i<ct;i++){
				zeros[i] = 0;
				ones[i] = 1;
			//	System.out.println(i + "\t" + varID2predField.get(i));
			}
			
			// obj function :- 0x1 + 0x2 + ... + xn = 0
			LinearObjectiveFunction f = 
				new LinearObjectiveFunction(zeros, 0);
			//LinearObjectiveFunction f = 
			//	new LinearObjectiveFunction(ones, 0);
			
			ArrayList<LinearConstraint> constraints = new ArrayList<LinearConstraint>();
			
			// constraint :- pi_1 + pi_2 + ... + pi_n >= 1
			constraints.add(new LinearConstraint(ones, Relationship.GEQ, 1));
			
			
			//TODO: Assume we know there are partitionings to some fields,
			// how can we push them back to other operators? (e.g., date -> d
			// throught aligned(.,.) relation).

			for(FelixClause c : op.allRelevantFelixClause){
								
				// for each clause, generate constraints for the linear system.
				ArrayList<Literal> allLiterals = c.getRegLiterals();
				
				for(Literal l1 : allLiterals){
					for(Literal l2 : allLiterals){
						if(l1.equals(l2)){
							continue;
						}
						
						if(op.outputPredicates.contains(l1.getPred()) == false ||
								op.outputPredicates.contains(l2.getPred()) == false){
							continue;
						}
						
						//System.err.println(l1 + "\t" + l2);
						
						HashSet<String> allVars = new HashSet<String>();
						allVars.addAll(l1.getVars());
						allVars.addAll(l2.getVars());
						
						for(String var : allVars){
							
							HashSet<Integer> shouldBeOne = new HashSet<Integer>();
							HashSet<Integer> shouldBeNegOne = new HashSet<Integer>();
							
							for(int i=0;i<l1.getVars().size();i++){
								if(var.equals(l1.getTerms().get(i).toString())){
							//		System.out.print("+" + toCanonicalFieldName(l1.getPred(), i) + " ");
									shouldBeOne.add(predField2varID.get(toCanonicalFieldName(l1.getPred(), i)));
								}
							}
							
							for(int i=0;i<l2.getVars().size();i++){
								if(var.equals(l2.getTerms().get(i).toString())){
							//		System.out.print("-" + toCanonicalFieldName(l2.getPred(), i) + " ");
									shouldBeNegOne.add(predField2varID.get(toCanonicalFieldName(l2.getPred(), i)));
								}
							}
							
							//System.out.println();
							
							if(shouldBeOne.contains(null) || shouldBeNegOne.contains(null)){
								continue;
							}
							
							double[] cons = new double[ct];
							int nOfPos = 0;
							int nOfNeg = 0;
							for(int i=0;i<ct;i++){
								cons[i] = 0;
								if(shouldBeOne.contains(i)){
									cons[i] += 1;
									nOfPos ++;
								}
								if(shouldBeNegOne.contains(i)){
									cons[i] += -1;
									nOfNeg ++;
								}
							}
							
							// [# appearance] x_i_in_left * pi = x_i_in_right * pi
							constraints.add(new LinearConstraint(cons, Relationship.EQ, 0));
						}
						
						
					}
				}
				//System.out.println(c);
			}
			
			RealPointValuePair solution = new 
			//SimplexSolver().optimize(f, constraints, 
			//		GoalType.MAXIMIZE, false);
			SimplexSolver().optimize(f, constraints, 
					GoalType.MAXIMIZE, false);
		
			// get the solution and translated them into
			// integer form
			double[] tmpss = solution.getPoint();
			ss = new double[tmpss.length];
			
			double largest = Double.MIN_VALUE;
			for(int i=0;i<tmpss.length;i++){
				if(tmpss[i] >= largest){
					largest = tmpss[i];
				}
			}

			for(int i=0;i<ss.length;i++){
				ss[i] = tmpss[i]/largest;
			}
			
			isDecomposable = true;

		}catch(Exception e){
		//	e.printStackTrace();
			isDecomposable = false;
		}
	}
	
	/**
	 * Return the GCD of the given two numbers.
	 * @param a
	 * @param b
	 * @return
	 */
	int gcd2(int a, int b){
		 if (b==0) return a;
		 return gcd2(b,a%b);
	}
	
	/**
	 * Return the GCD of the given multiple numbers.
	 * @param numbers
	 * @return
	 */
	int gcd(int... numbers){
		if(numbers.length == 1){
			return numbers[0];
		}
		int firstTwoGCD = gcd2(numbers[0], numbers[1]);
		int[] newNumbers = new int[numbers.length - 1];
		for(int i = 2; i<numbers.length; i++){
			newNumbers[i-2] = numbers[i];
		}
		newNumbers[newNumbers.length - 1] = firstTwoGCD;
		return gcd(newNumbers);
	}
	
	/**
	 * Given a clause appearing in the statOperator, return the appended expression
	 * to this clause to partition the data.
	 * @param fc
	 * @param _p if this value is not NULL and fc is NULL, generate a human-readable version for predicate _p
	 * @param base Number of partitions.
	 * @param nThread ID of the current partition.
	 * @param isSignature if this parameter is false, then it can be used as clause's constraints.
	 * @return
	 */
	public HashSet<Expression> getExpressions(FelixClause fc, Predicate _p, int base, int nThread, boolean isSignature){
		
		HashSet<String> processedExpSignature = new HashSet<String>();
		HashSet<Expression> ret = new HashSet<Expression>();
		
		if(fc != null){
			for(Literal l : fc.getRegLiterals()){
				
				Predicate p = l.getPred();
			
				Expression e = new Expression(Function.Eq);
				
				Expression emod = new Expression(Function.Modulo);
				
				Expression currentSumRoot = null;
				
				for(Integer fieldID : varID2pred.keySet()){
					
					if(ss[fieldID] == 0){
						continue;
					}
					
					if(!varID2pred.get(fieldID).equals(p)){
						continue;
					}
					
					if(currentSumRoot == null){
						Expression etmp = new Expression(Function.Multiply);
						if(isSignature == false){
							etmp.addArgument(l.getTerms().get(varID2field.get(fieldID)).isVariable() ? 
									Expression.exprVariableBinding(l.getTerms().get(varID2field.get(fieldID)).toString()) : 
									Expression.exprConstNum(l.getTerms().get(varID2field.get(fieldID)).constant() ));
						}else{
							etmp.addArgument(Expression.exprVariableBinding(l.getPred().getArgs().get(varID2field.get(fieldID)).toString()));
						}
						etmp.addArgument(Expression.exprConstNum(ss[fieldID]));
						
						currentSumRoot = etmp;
					}else{
						Expression etmp = new Expression(Function.Add);
						etmp.addArgument(currentSumRoot);
						
						Expression etoadd = new Expression(Function.Multiply);
						if(isSignature == false){
							etoadd.addArgument(l.getTerms().get(varID2field.get(fieldID)).isVariable() ? 
									Expression.exprVariableBinding(l.getTerms().get(varID2field.get(fieldID)).toString()) : 
										Expression.exprConstNum(l.getTerms().get(varID2field.get(fieldID)).constant() ));
						}else{
							etoadd.addArgument(Expression.exprVariableBinding(l.getPred().getArgs().get(varID2field.get(fieldID)).toString()));
						}
							
						etoadd.addArgument(Expression.exprConstNum(ss[fieldID]));
						
						etmp.addArgument(etoadd);
						
						currentSumRoot = etmp;
						
					}
				}
				
				if(currentSumRoot != null){
				
					Expression eround = new Expression(Function.Round);
					eround.addArgument(currentSumRoot);
					emod.addArgument(eround);
					emod.addArgument(Expression.exprConstInteger(base));
					e.addArgument(emod);
					e.addArgument(Expression.exprConstInteger(nThread));
					e.changeName = false;
					
					if(processedExpSignature.contains(e.toString())){
						continue;
					}
					
					processedExpSignature.add(e.toString());
					ret.add(e);
					
				}
			}
		}else if(_p != null){

			Expression e = new Expression(Function.Eq);
			
			Expression emod = new Expression(Function.Modulo);
			
			Expression currentSumRoot = null;
			
			for(Integer fieldID : varID2pred.keySet()){
				
				if(ss[fieldID] == 0){
					continue;
				}
				
				if(!varID2pred.get(fieldID).equals(_p)){
					continue;
				}
				
				if(currentSumRoot == null){
					Expression etmp = new Expression(Function.Multiply);
					if(isSignature == false){
						
					}else{
						etmp.addArgument(Expression.exprVariableBinding(_p.getArgs().get(varID2field.get(fieldID)).toString()));
					}
					etmp.addArgument(Expression.exprConstNum(ss[fieldID]));
					
					currentSumRoot = etmp;
				}else{
					Expression etmp = new Expression(Function.Add);
					etmp.addArgument(currentSumRoot);
					
					Expression etoadd = new Expression(Function.Multiply);
					if(isSignature == false){
						
					}else{
						etoadd.addArgument(Expression.exprVariableBinding(_p.getArgs().get(varID2field.get(fieldID)).toString()));
					}
						
					etoadd.addArgument(Expression.exprConstNum(ss[fieldID]));
					
					etmp.addArgument(etoadd);
					
					currentSumRoot = etmp;
					
				}
			}
			
			if(currentSumRoot != null){
			
				Expression eround = new Expression(Function.Round);
				eround.addArgument(currentSumRoot);
				emod.addArgument(eround);
				emod.addArgument(Expression.exprConstInteger(base));
				e.addArgument(emod);
				e.addArgument(Expression.exprConstInteger(nThread));
				e.changeName = false;
								
				processedExpSignature.add(e.toString());
				ret.add(e);
				
			}
		}
		return ret;
	}
	
}






