package felix.society;

import java.util.ArrayList;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

/**
 * A concurrent task.
 * @author czhang
 *
 */
public abstract class Task {

	/**
	 * Number of runs of this task.
	 */
	private int nRuns = 1;
	
	/**
	 * Type of task \in {TASKSET, TASKLIST, TASK}
	 * @author czhang
	 *
	 */
	public static enum TaskType {TASKSET, TASKLIST, TASK}
	
	/**
	 * A thread poll that only contains very very cheap threads (to
	 * support TaskSet > TaskList1, TaskList2 structure)
	 */
	public static ExecutorService softpool = Executors.newFixedThreadPool(1000);
	
	/**
	 * Type of this task.
	 */
	private TaskType type;
	
	/**
	 * If this task is a task list or task set, all subtasks.
	 */
	ArrayList<Task> subTasks = new ArrayList<Task>();
	
	/**
	 * What worker is currently running?
	 */
	public Worker currentWorker = null;
	
	/**
	 * What future is currently running (used for join).
	 */
	public Future currentFuture = null;
	
	/**
	 * The constructor.
	 * @param _type
	 */
	public Task(TaskType _type){
		type = _type;
	}
	
	/**
	 * 
	 * @return
	 */
	public TaskType getType(){
		return type;
	}
		
	public Task(TaskType _type, Integer _nRuns){
		type = _type;
		nRuns = _nRuns;
	}
	
	public void addSubTask(Task _task){
		this.subTasks.add(_task);
	}
	
	/**
	 * Execute this task.
	 * 
	 * @param pool The thread pool
	 * @throws Exception
	 */
	public void execute(ExecutorService pool) throws Exception{
		
		for(int i=0;i<nRuns;i++){
			if(this.type == TaskType.TASK){
				currentWorker = this.generateWorker();
				currentFuture = pool.submit(currentWorker);
				currentFuture.get();
			}else{
				currentWorker = this.generateWorker(pool);
				currentFuture = softpool.submit(currentWorker);
				currentFuture.get();
			}
		}
		
	}
	
	/**
	 * Generate worker for TASK (threads).
	 * @return
	 */
	public abstract Worker generateWorker();
	
	/**
	 * Generate worker for TASKLIST, TASKSET (threads).
	 * @param pool The thread pool in which subtasks will run.
	 * @return
	 */
	public abstract Worker generateWorker(ExecutorService pool);
	
	
	
	
}
