package felix.society;

import java.util.ArrayList;
import java.util.concurrent.ExecutorService;

import felix.society.Task.TaskType;

/**
 * TaskList -- Subtasks in it must be executed in order,
 * and the downstream tasks can be executed only if
 * the upstream tasks are joined.
 * @author czhang
 *
 */
public class TaskList extends Task{

	public TaskList(){
		super(TaskType.TASKLIST);
	}
	
	public TaskList(int _nRuns){
		super(TaskType.TASKLIST, _nRuns);
	}
	
	@Override
	public Worker generateWorker(){
		return null;
	}
	
	public Worker generateWorker(ExecutorService pool) {
		return new TaskListExecutionWorker(this.subTasks, pool);
	}

	/**
	 * Worker for running this TaskList.
	 * @author czhang
	 *
	 */
	public class TaskListExecutionWorker extends Worker{
		
		ArrayList<Task> subtasks;
		ExecutorService pool;
		
		public TaskListExecutionWorker(ArrayList<Task> _subtasks,
				ExecutorService _pool){
			subtasks = _subtasks;
			pool = _pool;
		}

		@Override
		public void run() {
			
			try{
				for(Task t : subTasks){
					if(t.getType() == TaskType.TASK){
						currentWorker = t.generateWorker();
						currentFuture = pool.submit(currentWorker);
						currentFuture.get();
					}else{
						currentWorker = t.generateWorker(pool);
						currentFuture = softpool.submit(currentWorker);
						currentFuture.get();
					}
				}
			}catch(Exception e){
				submitException(e);
			}
		}
		
	}

}










