package felix.task;

import java.io.BufferedReader;

import tuffy.util.DebugMan;
import tuffy.util.ExceptionMan;
import tuffy.util.FileMan;
import tuffy.util.UIMan;
import felix.dstruct.FelixQuery;
import felix.parser.FelixInputParser;
import felix.util.FelixConfig;

/**
 * Evidence loader that can run concurrently. 
 * It takes as input a filename and
 * a FelixQuery which it uses to parse the evidence file.
 * @author Ce Zhang
 *
 */
public class ConcurrentEvidenceLoader extends Thread{
	
	/**
	 * Path to the evidence file.
	 */
	String evidenceFileName = "";
	
	/**
	 * FelixQuery
	 */
	FelixQuery fq = null;
	
	/**
	 * The constructor.
	 * @param _fileName
	 * @param _fq
	 */
	public ConcurrentEvidenceLoader(String _fileName, FelixQuery _fq){
		this.evidenceFileName = _fileName;
		this.fq = _fq;
	}
	
	/**
	 * Parses evidence file
	 */
	public void run(){
		
		FelixInputParser parser = new FelixInputParser(this.fq);
		
		int chunkSize = FelixConfig.evidence_file_chunk_size;;
		String f = this.evidenceFileName;
		String g = FileMan.getGZIPVariant(f);
		if (g == null) {
			ExceptionMan.die("File does not exist: " + f);
		} else {
			f = g;
		}
		UIMan.println(">>> Parsing evidence file: " + f);
		if (FileMan.getFileSize(f) <= chunkSize) {
			parser.parseEvidenceFile(f);
		} else {
			try {
				long lineOffset = 0, lastChunkLines = 0;
				BufferedReader reader = FileMan.getBufferedReaderMaybeGZ(f);
				StringBuilder sb = new StringBuilder();
				String line = reader.readLine();
				int ct = 0;
				while (line != null) {
					
					ct ++;
					if(ct % 100000 == 0){
						UIMan.verbose(3, "LINE_" + ct);
					}
					
					sb.append(line);
					sb.append("\n");
					lastChunkLines++;
					if (sb.length() >= chunkSize) {
						parser.parseEvidenceString(sb.toString(),
								lineOffset);
						sb.delete(0, sb.length());
						sb = new StringBuilder();
						lineOffset += lastChunkLines;
						lastChunkLines = 0;
						UIMan.print(".");
					}
					line = reader.readLine();
				}
				reader.close();
				if (sb.length() > 0) {
					parser.parseEvidenceString(sb.toString(), lineOffset);
				}
				UIMan.println();
			} catch (Exception e) {
				ExceptionMan.handle(e);
			}
		}
		try {
			DebugMan.runGC();
		} catch (Exception e) {
			e.printStackTrace();
		}
		
	}
	

}
