package felix.task;

import java.io.FileInputStream;
import java.util.concurrent.ExecutorService;

import org.postgresql.PGConnection;

import tuffy.util.Timer;
import tuffy.util.UIMan;

import felix.dstruct.DataMovementOperator;
import felix.dstruct.StatOperator;
import felix.optimizer.DMOOptimizer;
import felix.society.Task;
import felix.society.Worker;
import felix.society.Task.TaskType;

/**
 * {@link Task} Object for executing an operator.
 * @author czhang
 *
 */
public class ExecuteOperatorTask extends Task{

	/**
	 * Operator.
	 */
	public StatOperator sop;
	
	/**
	 * Constructor.
	 * @param _sop
	 */
	public ExecuteOperatorTask(StatOperator _sop){
		super(TaskType.TASK);
		sop = _sop;
	}
	
	@Override
	public Worker generateWorker() {
		return new ExecuteOperatorWorker(sop);
	}
	
	/**
	 * A worker that simply calls the run() function
	 * of the operator object.
	 * @author czhang
	 *
	 */
	public class ExecuteOperatorWorker extends Worker{

		public StatOperator sop;
		public DMOOptimizer dmoo;
		
		public ExecuteOperatorWorker(StatOperator _sop){
			super();
			sop = _sop;
		}
		
		@Override
		public void run() {
			try{
				
				sop.run();
				
			}catch(Exception e){
				submitException(e);
			}
		}
	}

	@Override
	public Worker generateWorker(ExecutorService pool) {
		return null;
	}
	

}
