package felix.task;

import java.io.FileInputStream;
import java.util.concurrent.ExecutorService;

import org.postgresql.PGConnection;

import tuffy.util.Timer;
import tuffy.util.UIMan;

import felix.dstruct.DataMovementOperator;
import felix.dstruct.StatOperator;
import felix.optimizer.DMOOptimizer;
import felix.society.Task;
import felix.society.Worker;
import felix.society.Task.TaskType;

/**
 * {@link Task} Object for optimizing a {@link DataMovementOperator}.
 * @author czhang
 *
 */
public class OptimizeDMOTask extends Task{

	/**
	 * Operator containing the DMO.
	 */
	public StatOperator sop;
	
	/**
	 * DMO to be optimized.
	 */
	public DMOOptimizer dmoo;
	
	/**
	 * Constructor.
	 * @param _sop
	 * @param _dmoo
	 */
	public OptimizeDMOTask(StatOperator _sop, DMOOptimizer _dmoo){
		super(TaskType.TASK);
		sop = _sop;
		dmoo = _dmoo;
	}
	
	@Override
	public Worker generateWorker() {
		return new OptimizeDMOWorker(sop, dmoo);
	}
	
	/**
	 * A worker that optimize the DMO.
	 * @author czhang
	 *
	 */
	public class OptimizeDMOWorker extends Worker{

		public StatOperator sop;
		public DMOOptimizer dmoo;
		
		public OptimizeDMOWorker(StatOperator _sop, DMOOptimizer _dmoo){
			super();
			sop = _sop;
			dmoo = _dmoo;
		}
		
		@Override
		public void run() {
			try{
								
				Timer.start("Mat-Timer");
				System.out.println(">>> Optimize DMOs for " + sop + "... ");
				sop.prepare();
				
				for(DataMovementOperator dmo : sop.getAllDMOs()){
					
					//if(!dmo.hasPrepared){
						if(dmo.allowOptimization == false){
							dmo.copyLogicPlanToPhysicalPlanDirectly();
						}else{
							//optimize the materialization trade-off.
							dmoo.optimizeMateralization(dmo);	
						}
						dmo.prepare();

				}
				System.out.println("Uses " + Timer.elapsed("Mat-Timer") + "\n");
		
				
			}catch(Exception e){
				submitException(e);
			}
		}
	}

	@Override
	public Worker generateWorker(ExecutorService pool) {
		return null;
	}
	

}
