package felix.test;

import static org.junit.Assert.*;

import java.io.BufferedReader;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.junit.Test;

import felix.main.Felix;
import felix.parser.FelixCommandOptions;
import felix.util.FelixConfig;
import felix.util.FelixUIMan;


/**
 * Test CRF operator's quality.
 * @author Ce Zhang
 *
 */
public class CRFTest{
	
	/**
	 * Compare Marginal inference results with CRF++ on a real world data set.
	 */
	@Test
	public final void testCRFMarginal() {
		try{
			String[] args = {"-e", "test/felix/crf_test/evidence_small.db", "-i", "test/felix/crf_test/prog.mln",
					"-o", "test/testOutput.txt_label", "-queryFile", "test/felix/crf_test/query.db", 
					"-marginal", "-verbose", "2", "-threads", "10"};
			
			for(int wwww = 0; wwww < 2; wwww++){
				
				FelixConfig.overrideID();
				FelixCommandOptions options = FelixUIMan.parseCommand(args);
				
				if(wwww == 1){
					options.useDualDecomposition = true;
					FelixConfig.nDDIT = 2;
				}
				
				
				new Felix().run(options);
				
				HashMap<String, Double> gt = new HashMap<String, Double>();
				
				BufferedReader br = new BufferedReader(new FileReader("test/felix/crf_test/gt_v1"));
				Pattern p = Pattern.compile("gt\\((.*?), (.*?), \"(.*?)/(.*?)\"\\)");
				String line;
				while((line = br.readLine()) != null){
					
					Matcher m = p.matcher(line);
					if(m.find()){
						gt.put(m.group(1) + "|" + m.group(2) + "|" + m.group(3), Double.valueOf(m.group(4)));
					}				
				}
				
				String outfile = "test/testOutput.txt_label";
				
				if(wwww == 1){
					outfile = outfile + "_snap_0";
				}
				
				br = new BufferedReader(new FileReader(outfile));
				p = Pattern.compile("(.*?)\tlabel\\(\"(.*?)\", \"(.*?)\", \"(.*?)\"\\)");
				int ct = 0;
				
				while((line = br.readLine()) != null){
					
					Matcher m = p.matcher(line);
					if(m.find()){
						String key = m.group(2) + "|" + m.group(3);
						Double ass = Double.valueOf(m.group(1));
						if(m.group(4).equals("WIN")){
							String a = key + "|Winner";
							if(gt.containsKey(a)){
								
								Double target = gt.get(a);
								if(Math.abs(ass - target) < 0.1){
									ct++;
								}
							}
						}else if(m.group(4).equals("LOS")){
							String a = key + "|Loser";
							if(gt.containsKey(a)){
								Double target = gt.get(a);
								if(Math.abs(ass - target) < 0.1){
									ct++;
								}
							}						
						}else if(m.group(4).equals("NON")){
							String a = key + "|--";
							if(gt.containsKey(a)){
								Double target = gt.get(a);
								if(Math.abs(ass - target) < 0.1){
									ct++;
								}
							}
	
						}
					}
					
				}
				System.out.println("\n\n# OF LABELS = " +ct + "\n\n");
				assertTrue(ct > 9200);
			
			}
			
		}catch(Exception e){
			e.printStackTrace();
			assertTrue(false);
		}
	}
	
	/**
	 * Compare MAP inference results with CRF++ on a real world data set.
	 */
	@Test
	public final void testCRFMAP() {
		try{
			String[] args = {"-e", "test/felix/crf_test/evidence_small.db", "-i", "test/felix/crf_test/prog.mln",
					"-o", "test/testOutput.txt_label", "-queryFile", "test/felix/crf_test/query.db", "-verbose", "2"
					};
			
			for(int wwww = 0; wwww < 2; wwww++){
			
				FelixConfig.overrideID();
				FelixCommandOptions options = FelixUIMan.parseCommand(args);
				
				if(wwww == 1){
					options.useDualDecomposition = true;
					FelixConfig.nDDIT = 2;
				}
				
				new Felix().run(options);
				
				String outfile = "test/testOutput.txt_label";
				
				if(wwww == 1){
					outfile = outfile + "_snap_0";
				}
				
				HashMap<String, String> gt = new HashMap<String, String>();
				
				BufferedReader br = new BufferedReader(new FileReader("test/felix/crf_test/gt"));
				Pattern p = Pattern.compile("gt\\((.*?), (.*?), \"(.*?)\"\\)");
				String line;
				while((line = br.readLine()) != null){
					
					Matcher m = p.matcher(line);
					if(m.find()){
						gt.put(m.group(1) + "|" + m.group(2), m.group(3));
					}				
				}
				
				
				br = new BufferedReader(new FileReader(outfile));
				p = Pattern.compile("label\\(\"*(.*?)\"*, \"*(.*?)\"*, \"(.*?)\"\\)");
				int ct = 0;
				
				while((line = br.readLine()) != null){
					
					Matcher m = p.matcher(line);
					if(m.find()){
						String key = m.group(1) + "|" + m.group(2);
						if(m.group(3).equals("WIN")){
							
							if(!gt.get(key).equals("Winner")){
								System.out.println(line);
							}else{
								ct++;
							}
						}else if(m.group(3).equals("LOS")){
							
							if(!gt.get(key).equals("Loser")){
								System.out.println(line);
							}else{
								ct++;
							}					
						}else if(m.group(3).equals("NON")){
							if(!gt.get(key).equals("--")){
								System.out.println(line);
							}else{
								ct++;
							}
						}
					}
					
				}
				
				assertTrue(ct > 9200);
				System.out.println("\n\n# OF LABELS = " +ct + "\n\n");
			}
			
			
		}catch(Exception e){
			e.printStackTrace();
			assertTrue(false);
		}
	}
	
	
	
}






