package felix.test;

import static org.junit.Assert.*;

import java.io.BufferedReader;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.junit.BeforeClass;
import org.junit.Test;

import tuffy.util.UIMan;

import felix.main.Felix;
import felix.parser.FelixCommandOptions;
import felix.util.FelixConfig;
import felix.util.FelixUIMan;


/**
 * Test LR operator's quality.
 * @author Ce Zhang
 *
 */
public class DDTest{
	
	public static ArrayList<Double> planCosts = new ArrayList<Double>();
	public static String exeTime = "";
	

	/**
	 * Test Marginal inference results of LR on a test case with manually-built ground truth.
	 */
	@Test
	public final void testDDMarginal() {
		try{
			String[] args = {"-e", "test/felix/DDTest/evidence.db", "-i", "test/felix/DDTest/prog.mln",
					 "-o", "test/testOutput.txt", "-q", "op1", "-marginal", "-dd", "-nDD", "5"
					 , "-mcsatSamples", "1000"};
			
			
			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
			
			
			new Felix().run(options);
			
			String outfile = "test/testOutput.txt";

		
			BufferedReader br = new BufferedReader(new FileReader("test/testOutput.txt"));
			
			Pattern p1 = Pattern.compile("(.*?)\top1\\(\"K2\", \"L2\"\\)");
			Pattern p2 = Pattern.compile("(.*?)\top1\\(\"K1\", \"L1\"\\)");	
			
			String line;
			while( (line = br.readLine()) != null){
				int ct = 0;
				
				Matcher m = p1.matcher(line);
				if(m.find() && Double.valueOf(m.group(1)) > 0.6){
					ct++;
				}
				
				m = p2.matcher(line);
				if(m.find() && Double.valueOf(m.group(1)) < 0.3){
					ct++;
				}
				
				
				UIMan.warn(""+ct);
				assertTrue(ct == 1);
				
				
			}
			
		}catch(Exception e){
			e.printStackTrace();
		}
	}
	
	/**
	 * Test MAP inference results of LR on a test case with manually-built ground truth.
	 */
	@Test
	public final void testDDMAP() {
		try{
			String[] args = {"-e", "test/felix/DDTest/evidence.db", "-i", "test/felix/DDTest/prog.mln",
					 "-o", "test/testOutput.txt", "-q", "op1", "-dd", "-nDD", "10"
					 , "-maxFlips", "10000000"};
			
			
			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
			
			
			new Felix().run(options);
			
			String outfile = "test/testOutput.txt";

		
			BufferedReader br = new BufferedReader(new FileReader("test/testOutput.txt"));
			
			Pattern p1 = Pattern.compile("top1\\(\"K1\", \"L2\"\\)");
			Pattern p2 = Pattern.compile("top1\\(\"K2\", \"L2\"\\)");
			
			
			String line;
			while( (line = br.readLine()) != null){
				int ct = 0;
				
				System.out.println(line);
				
				if(line.equals("op1(\"K2\", \"L2\")") || line.equals("op1(\"K1\", \"L2\")")){
					assertTrue(true);
				}else{
					assertTrue(false);
				}
			
			}
			
		}catch(Exception e){
			e.printStackTrace();
		}
	}
	
	
	
}






