package felix.test;

import static org.junit.Assert.*;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.util.ArrayList;

import org.junit.BeforeClass;
import org.junit.Test;

import tuffy.db.RDB;
import tuffy.util.Config;
import tuffy.util.FileMan;
import tuffy.util.UIMan;

import felix.main.Felix;
import felix.parser.FelixCommandOptions;
import felix.util.FelixConfig;
import felix.util.FelixUIMan;

/**
 * Test using evidence supplied in existing database relations.
 * @author Josh Slauson
 *
 */
public class IODBTest {
	
	static RDB db;
	
	/**
	 * Perform required setup for this test.
	 */
	@BeforeClass
	public static final void setUp() {
		// delete existing output files
		FileMan.removeFile("test/felix/output/out.txt_IODBTest_simpleTest");
		FileMan.removeFile("test/felix/output/out.txt_IODBTest_comboTest");
		FileMan.removeFile("test/felix/output/out.txt_IODBTest_emptyTable");
		FileMan.removeFile("test/felix/output/out.txt_IODBTest_incorrectMissingTable");
		FileMan.removeFile("test/felix/output/out.txt_IODBTest_incorrectColumnName");
		FileMan.removeFile("test/felix/output/out.txt_IODBTest_incorrectMissingIndices");
		FileMan.removeFile("test/felix/output/out.txt_IODBTest_incorrectMissingTruthPrior");
		
		FelixConfig.overrideID();
		
		UIMan.parseConfigFile(Config.path_conf);
		
		if(FelixConfig.evidDBSchema != null){
			FelixConfig.db_schema = FelixConfig.evidDBSchema;
		}

		db = RDB.getRDBbyConfig();
		db.schema = FelixConfig.db_schema;
		db.resetSchema(FelixConfig.db_schema);
		
		if(!db.isSchemaExists("felix_test")) {
			String query = "CREATE SCHEMA felix_test";
			db.execute(query);
		}
		
		db.dropTable("felix_test.evid");
		String query = "CREATE TABLE felix_test.evid (truth boolean, prior integer, a1 text, b2 text)";
		db.execute(query);
		
		query = "INSERT INTO felix_test.evid VALUES (true, 0, 1, 2), (true, 0, 4, 5)";
		db.execute(query);		

	}
	
	/**
	 * Checks if output file contains expected results.
	 * @param output
	 * @param expect
	 */
	public void checkOutput(String output, String[] expect) {
		ArrayList<String> lines = FileMan.getLines(output);
		System.out.println("Checking output size: " + lines.size() + " = " + expect.length/2 + "?");
		assertTrue(lines.size() == expect.length/2);
		
		for(int i = 0; i < lines.size(); i++) {
			System.out.println("Checking output: " + lines.get(i));

			boolean match = false;
			for(int j = 0; j + 1 < expect.length; j+=2) {
				if(lines.get(i).equals("fact(\"" + expect[j] + "\", \"" + expect[j+1] + "\")")) {
					match = true;
				}
			}
			if(!match) {
				System.out.println("\tNo match");
			}
			assertTrue(match);
		}

	}

	/**
	 * Simple smoke test.
	 */
	@Test
	public void simpleTest() {
		try {
			
			String output = "test/felix/output/out.txt_IODBTest_simpleTest";
			String[] args = {"-i", "test/felix/io_test/prog.mln-db", "-o", 
					"test/felix/output/out.txt_IODBTest_simpleTest", "-queryFile", "test/felix/io_test/query.db", "-keepData"};
						
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
			new Felix().run(options);
			
			String[] expect = {"1", "2", "4", "5"};
			checkOutput(output, expect);
			
		} catch(Exception e) {
			e.printStackTrace();
			assertTrue(false);
		}
	}
	
	/**
	 * Combo test using evidence from a file and evidence from a database table.
	 */
	@Test
	public void comboTest() {
		try {
			
			String output = "test/felix/output/out.txt_IODBTest_comboTest"; 
			String[] args = {"-e", "test/felix/io_test/evid2.db", "-i", "test/felix/io_test/prog2.mln-db", "-o", 
					output, "-queryFile", "test/felix/io_test/query.db", "-keepData"};
						
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
			new Felix().run(options);
			
			String[] expect = {"1", "2", "4", "5", "Hello", "World", "UW", "Madison"};
			checkOutput(output, expect);
			
		} catch(Exception e) {
			e.printStackTrace();
			assertTrue(false);
		}
	}
	
//	@Test
//	public void emptyTable() {
//		try {
//			
//			// Create evidence table
//			db.dropTable("felix_test.evid");
//			String query = "CREATE TABLE felix_test.evid (truth boolean, prior integer, a1 text, b2 text)";
//			db.execute(query);
//			
//			String output = "test/felix/output/out.txt_IODBTest_emptyTable";
//			String[] args = {"-i", "test/felix/io_test/prog.mln-db", "-o", output, 
//					"-queryFile", "test/felix/io_test/query.db", "-keepData"};
//						
//			FelixCommandOptions options = FelixUIMan.parseCommand(args);
//			new Felix().run(options);
//			
//			checkOutput(output);
//
//		} catch(Exception e) {
//			e.printStackTrace();
//			assertTrue(false);
//		}
//	}
	
	// TODO: fix issue, then uncomment
//		@Test
//		public void incorrectMissingTable() {
//			try {
//				
//				db.dropTable("felix_test.evid");
//				
//				String output = "test/felix/output/out.txt_IODBTest_incorrectMissingTable";
//				String[] args = {"-i", "test/felix/io_test/prog.mln-db", "-o", output,
//						"-queryFile", "test/felix/io_test/query.db", "-keepData"};
//							
//				FelixCommandOptions options = FelixUIMan.parseCommand(args);
//				new Felix().run(options);
//				
//				checkOutput(output);
//
//			} catch(Exception e) {
//				e.printStackTrace();
//				assertTrue(false);
//			}
//		}
	
	// TODO: fix issue, then uncomment
//	@Test
//	public void incorrectColumnName() {
//		try {
//			
//			// Create evidence table
//			db.dropTable("felix_test.evid");
//			String query = "CREATE TABLE felix_test.evid (truth boolean, prior integer, a1 text, a2 text)";
//			db.execute(query);
//			
//			query = "INSERT INTO felix_test.evid VALUES (true, 0, 1, 2), (true, 0, 4, 5)";
//			db.execute(query);
//			
//			String output = "test/felix/output/out.txt_IODBTest_incorrectColumnName";
//			String[] args = {"-i", "test/felix/io_test/prog.mln-db", "-o", output,
//					"-queryFile", "test/felix/io_test/query.db", "-keepData"};
//						
//			FelixCommandOptions options = FelixUIMan.parseCommand(args);
//			new Felix().run(options);
//			
//			checkOutput(output);
//
//		} catch(Exception e) {
//			e.printStackTrace();
//			assertTrue(false);
//		}
//	}


	// TODO: fix issue, then uncomment
//	@Test
//	public void incorrectMissingIndices() {
//		try {
//			
//			// Create evidence table
//			db.dropTable("felix_test.evid.evid");
//			String query = "CREATE TABLE felix_test.evid (truth boolean, prior integer, a text, b text)";
//			db.execute(query);
//			
//			query = "INSERT INTO felix_test.evid VALUES (true, 0, 1, 2), (true, 0, 4, 5)";
//			db.execute(query);
//			
//			String output = "test/felix/output/out.txt_IODBTest_incorrectMissingIndices";
//			String[] args = {"-i", "test/felix/io_test/prog.mln-db", "-o", output,
//					"-queryFile", "test/felix/io_test/query.db", "-keepData"};
//						
//			FelixCommandOptions options = FelixUIMan.parseCommand(args);
//			new Felix().run(options);
//			
//			checkOutput(output);
//
//		} catch(Exception e) {
//			e.printStackTrace();
//			assertTrue(false);
//		}
//	}

	// TODO: fix issue, then uncomment
//	@Test
//	public void incorrectMissingTruthPrior() {
//		try {
//			
//			// Create evidence table
//			db.dropTable("felix_test.evid");
//			String query = "CREATE TABLE felix_test.evid (a1 text, b2 text)";
//			db.execute(query);
//			
//			query = "INSERT INTO felix_test.evid VALUES (1, 2), (4, 5)";
//			db.execute(query);
//			
//			String output = "test/felix/output/out.txt_IODBTest_incorrectMissingTruthPrior";
//			String[] args = {"-i", "test/felix/io_test/prog.mln-db", "-o", output,
//					"-queryFile", "test/felix/io_test/query.db", "-keepData"};
//						
//			FelixCommandOptions options = FelixUIMan.parseCommand(args);
//			new Felix().run(options);
//			
//			checkOutput(output);
//
//		} catch(Exception e) {
//			e.printStackTrace();
//			assertTrue(false);
//		}
//	}

}
