package felix.test;

import static org.junit.Assert.*;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;

import org.junit.BeforeClass;
import org.junit.Test;

import tuffy.db.RDB;
import tuffy.util.Config;
import tuffy.util.FileMan;
import tuffy.util.UIMan;

import felix.main.Felix;
import felix.parser.FelixCommandOptions;
import felix.util.FelixConfig;
import felix.util.FelixUIMan;

/**
 * Test for generating evidence through the Blah Blah feature extraction language.
 * @author Josh Slauson
 *
 */
public class IOHDFSTest {

	/*

	NOTE: these tests assume files hdfs://d-02.cs.wisc.edu:9000/temp/test.txt, hdfs://d-02.cs.wisc.edu:9000/temp/test.xml exist
	 
	test.txt contents:
		This is a test.
		This is another sentence.
		This is the end.
	
 	test.xml contents:
		<DOC>
		<HEADLINE>
		This is the headline.
		</HEADLINE>
		<POST>
		This is a test.
		</POST>
		<POST>
		This is another sentence.
		</POST>
		<POST>
		This is the end.
		</POST>
		</DOC>

	 */

	// TODO: fix "Failed to connect to PostgreSQL! FATAL: sorry, too many clients already"
	// maybe do init db, db.close() between tests?
 
	// TODO: run tests without "-local"
	
	public static RDB db;
	
	/**
	 * Perform required setup for this test.
	 */
	@BeforeClass
	public static final void setUp() {
		// delete existing output files
		FileMan.removeFile("test/felix/output/out.txt_IOHDFSTest_simpleTest");
		FileMan.removeFile("test/felix/output/out.txt_IOHDFSTest_initTest");
		FileMan.removeFile("test/felix/output/out.txt_IOHDFSTest_importTest");
		FileMan.removeFile("test/felix/output/out.txt_IOHDFSTest_reduceTest");
		FileMan.removeFile("test/felix/output/out.txt_IOHDFSTest_simpleTestXML");
		FileMan.removeFile("test/felix/output/out.txt_IOHDFSTest_initTestXML");
		FileMan.removeFile("test/felix/output/out.txt_IOHDFSTest_importTestXML");
		FileMan.removeFile("test/felix/output/out.txt_IOHDFSTest_reduceTestXML");
		FileMan.removeFile("test/felix/output/out.txt_IOHDFSTest_localTest");
		FileMan.removeFile("test/felix/output/out.txt_IOHDFSTest_comboTest1");
		FileMan.removeFile("test/felix/output/out.txt_IOHDFSTest_comboTest2");
		FileMan.removeFile("test/felix/output/out.txt_IOHDFSTest_comboTest3");
		FileMan.removeFile("test/felix/output/out.txt_IOHDFSTest_comboTest4");
		FileMan.removeFile("test/felix/output/out.txt_IOHDFSTest_repeatTest");
		FileMan.removeFile("test/felix/output/out.txt_IOHDFSTest_xmlNoMatchTest");
		FileMan.removeFile("test/felix/output/out.txt_IOHDFSTest_commentTest");
		FileMan.removeFile("test/felix/output/out.txt_IOHDFSTest_printTest");


		FelixConfig.overrideID();
		
		UIMan.parseConfigFile(Config.path_conf);
		
		if(FelixConfig.evidDBSchema != null){
			FelixConfig.db_schema = FelixConfig.evidDBSchema;
		}

		db = RDB.getRDBbyConfig();
		db.schema = FelixConfig.db_schema;
		db.resetSchema(FelixConfig.db_schema);
		
		if(!db.isSchemaExists("felix_test")) {
			String query = "CREATE SCHEMA felix_test";
			db.execute(query);
		}

		db.dropTable("felix_test.evid");
		String query = "CREATE TABLE felix_test.evid (truth boolean, prior integer, a1 text, b2 text)";
		db.execute(query);
		
		query = "INSERT INTO felix_test.evid VALUES (true, 0, 11, 12), (true, 0, 14, 15)";
		db.execute(query);		
	}
	
	/**
	 * Check if evidence was correctly loaded into a database table.
	 * @param expect
	 */
	public void checkEvidence(String[] expect) {
		// Check if evidence table exists
		db = RDB.getRDBbyConfig(FelixConfig.db_schema);
		long count = db.countTuples("felix_test._fe_raw_length");
		System.out.println("Checking evidence count: " + count + " = " + expect.length + "?");
		assertTrue(count == expect.length);
		
		// Check evidence table contents
		String query = "SELECT * FROM felix_test._fe_raw_length";
		ResultSet rs = db.query(query);
		
		try {
			while(rs != null && rs.next()) {
				String text = rs.getString("text1");
				String len = rs.getString("len2");
				
				System.out.println("Checking evidence contents: " + text + " (" + len + ")");
				
				boolean match = false;
				for(int i = 0; i < expect.length; i++) {
					if(text.equals(expect[i])) {
						System.out.println("\tChecking evience length: " + len + " = " + expect[i].length() + "?");
						assertTrue(len.equals("" + expect[i].length()));
						match = true;
					}
				}
				if(!match) {
					System.out.println("\tNo match");
				}
				assertTrue(match);
			}
			
			db.close();
			
		} catch (SQLException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			assertTrue(false);
		}

	}

	/**
	 * Checks if output file contains expected results.
	 * @param output
	 * @param expect
	 */
	public void checkOutput(String output, String[] expect) {
		ArrayList<String> lines = FileMan.getLines(output);
		System.out.println("Checking output size: " + lines.size() + " = " + expect.length + "?");
		assertTrue(lines.size() == expect.length);
		
		for(int i = 0; i < lines.size(); i++) {
			System.out.println("Checking output: " + lines.get(i));

			boolean match = false;
			for(int j = 0; j < expect.length; j++) {
				if(lines.get(i).equals("long(\"" + expect[j] + "\")")) {
					match = true;
				}
			}
			if(!match) {
				System.out.println("\tNo match");
			}
			assertTrue(match);
		}

	}

	/**
	 * Simple smoke test.
	 */
	@Test
	public void simpleTest() {
		try {
			String output = "test/felix/output/out.txt_IOHDFSTest_simpleTest";
			String[] args = {"-i", "test/felix/io_test/prog.mln-hdfs-simple,test/felix/io_test/prog.mln-hdfs", "-o", output, 
					"-queryFile", "test/felix/io_test/query.db-hdfs", "-keepData", "-auxSchema",
					"felix_test", "-local"};
			
			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
			new Felix().run(options);
			
			String[] expect = { "This is a test.", "This is another sentence.", "This is the end."};
			checkEvidence(expect);
			checkOutput(output, expect);

		} catch(InterruptedException e) {
			e.printStackTrace();
			assertTrue(false);
		}
	}

	/**
	 * Test using @MAPINIT.
	 */
	@Test
	public void initTest() {
		try {
			String output = "test/felix/output/out.txt_IOHDFSTest_initTest";
			String[] args = {"-i", "test/felix/io_test/prog.mln-hdfs-init,test/felix/io_test/prog.mln-hdfs", "-o", output, 
					"-queryFile", "test/felix/io_test/query.db-hdfs", "-keepData", "-auxSchema",
					"felix_test", "-local"};
			
			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
			new Felix().run(options);
			
			String[] expect = { "headThis is a test.", "headThis is another sentence.", "headThis is the end."};
			checkEvidence(expect);
			checkOutput(output, expect);
		} catch(InterruptedException e) {
			e.printStackTrace();
			assertTrue(false);
		}
	}
	
	/**
	 * Test using an import statement.
	 */
	@Test
	public void importTest() {
		try {
			String output = "test/felix/output/out.txt_IOHDFSTest_importTest";
			String[] args = {"-i", "test/felix/io_test/prog.mln-hdfs-import,test/felix/io_test/prog.mln-hdfs", "-o", output, 
					"-queryFile", "test/felix/io_test/query.db-hdfs", "-keepData", "-auxSchema",
					"felix_test", "-local"};
			
			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
			new Felix().run(options);
			
			String[] expect = { "This is a test", "This is another sentence", "This is the end"};
			checkEvidence(expect);
			checkOutput(output, expect);
		} catch(InterruptedException e) {
			e.printStackTrace();
			assertTrue(false);
		}
	}
	
	/**
	 * Test using a custom @REDUCE method.
	 */
	@Test
	public void reduceTest() {
		/*
		try {
			String output = "test/felix/output/out.txt_IOHDFSTest_reduceTest";
			String[] args = {"-i", "test/felix/io_test/prog.mln-hdfs-reduce", "-o", output, 
					"-queryFile", "test/felix/io_test/query.db-hdfs", "-keepData", "-auxSchema",
					"felix_test", "-local"};
			
			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
			new Felix().run(options);
			
			String[] expect = { "This is a test.", "This is another sentence.", "This is the end.", "This is a test.", "This is another sentence.", "This is the end." };
			checkEvidence(expect);
			checkOutput(output, expect);
		} catch(InterruptedException e) {
			e.printStackTrace();
			assertTrue(false);
		}
		*/
	}

	/**
	 * Simple smoke test with an XML file.
	 */
	@Test
	public void simpleTestXML() {
		try {
			String output = "test/felix/output/out.txt_IOHDFSTest_simpleTestXML";
			String[] args = {"-i", "test/felix/io_test/prog.mln-hdfs-simple-xml,test/felix/io_test/prog.mln-hdfs", "-o", output, 
					"-queryFile", "test/felix/io_test/query.db-hdfs", "-keepData", "-auxSchema",
					"felix_test", "-local"};
			
			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
			new Felix().run(options);
			
			String[] expect = { "This is a test.", "This is another sentence.", "This is the end."};
			checkEvidence(expect);
			checkOutput(output, expect);
		} catch(InterruptedException e) {
			e.printStackTrace();
			assertTrue(false);
		}
	}
	
	/**
	 * Test using @MAPINIT with with an XML file.
	 */
	@Test
	public void initTestXML() {
		try {
			String output = "test/felix/output/out.txt_IOHDFSTest_initTestXML";
			String[] args = {"-i", "test/felix/io_test/prog.mln-hdfs-init-xml,test/felix/io_test/prog.mln-hdfs", "-o", output, 
					"-queryFile", "test/felix/io_test/query.db-hdfs", "-keepData", "-auxSchema",
					"felix_test", "-local"};
			
			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
			new Felix().run(options);
			
			String[] expect = { "headThis is a test.", "headThis is another sentence.", "headThis is the end."};
			checkEvidence(expect);
			checkOutput(output, expect);
		} catch(InterruptedException e) {
			e.printStackTrace();
			assertTrue(false);
		}
	}
	
	/**
	 * Test using an import statement with an XML file.
	 */
	@Test
	public void importTestXML() {
		try {
			String output = "test/felix/output/out.txt_IOHDFSTest_importTestXML";
			String[] args = {"-i", "test/felix/io_test/prog.mln-hdfs-import-xml,test/felix/io_test/prog.mln-hdfs", "-o", output, 
					"-queryFile", "test/felix/io_test/query.db-hdfs", "-keepData", "-auxSchema",
					"felix_test", "-local"};
			
			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
			new Felix().run(options);
			
			String[] expect = { "This is a test", "This is another sentence", "This is the end"};
			checkEvidence(expect);
			checkOutput(output, expect);
		} catch(InterruptedException e) {
			e.printStackTrace();
			assertTrue(false);
		}
	}
	
	/**
	 * Test using a custom @REDUCE method with an XML file.
	 */
	@Test
	public void reduceTestXML() {
		/*
		try {
			String output = "test/felix/output/out.txt_IOHDFSTest_reduceTestXML";
			String[] args = {"-i", "test/felix/io_test/prog.mln-hdfs-reduce-xml", "-o", output, 
					"-queryFile", "test/felix/io_test/query.db-hdfs", "-keepData", "-auxSchema",
					"felix_test", "-local"};
			
			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
			new Felix().run(options);
			
			String[] expect = { "This is a test.", "This is another sentence.", "This is the end.", "This is a test.", "This is another sentence.", "This is the end." };
			checkEvidence(expect);
			checkOutput(output, expect);
		} catch(InterruptedException e) {
			e.printStackTrace();
			assertTrue(false);
		}
		*/
	}
	
	/**
	 * Test using -local option.
	 */
	@Test
	public void localTest() {
		try {
			String output = "test/felix/output/out.txt_IOHDFSTest_localTest";
			String[] args = {"-i", "test/felix/io_test/prog.mln-hdfs-local,test/felix/io_test/prog.mln-hdfs", "-o", output, 
					"-queryFile", "test/felix/io_test/query.db-hdfs", "-keepData", "-auxSchema",
					"felix_test", "-local"};
			
			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
			new Felix().run(options);
			
			// TODO: check table for features
			
			String[] expect = { "This is a test.", "This is another sentence.", "This is the end."};
			checkEvidence(expect);
			checkOutput(output, expect);
		} catch(InterruptedException e) {
			e.printStackTrace();
			assertTrue(false);
		}
	}
	
	/**
	 * Combo test using evidence generated with Blah Blah and evidence from a file.
	 */
	@Test
	public void comboTest1() {
		try {
			String output = "test/felix/output/out.txt_IOHDFSTest_comboTest1";
			String[] args = {"-e", "test/felix/io_test/evid.db", "-i", 
					"test/felix/io_test/prog.mln-hdfs-combo1,test/felix/io_test/prog.mln-hdfs", "-o", output, 
					"-queryFile", "test/felix/io_test/query.db-hdfs", "-keepData", "-auxSchema",
					"felix_test", "-q", "fact", "-local"};
			
			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
			new Felix().run(options);
			
			String[] expect = { "This is a test.", "This is another sentence.", "This is the end."};
			checkEvidence(expect);

			ArrayList<String> lines = FileMan.getLines(output);
			assertTrue(lines.size() == 5);
			for(int i = 0; i < lines.size(); i++) {
				assertTrue(lines.get(i).equals("long(\"This is a test.\")") || 
						lines.get(i).equals("long(\"This is another sentence.\")") ||
						lines.get(i).equals("long(\"This is the end.\")") ||
						lines.get(i).equals("fact(\"1\", \"2\")") || 
						lines.get(i).equals("fact(\"4\", \"5\")"));
			}

		} catch(InterruptedException e) {
			e.printStackTrace();
			assertTrue(false);
		}
	}

	/**
	 * Combo test using evidence generated with Blah Blah and evidence from a database table.
	 */
	@Test
	public void comboTest2() {
			try {
			String output = "test/felix/output/out.txt_IOHDFSTest_comboTest2";
			String[] args = {"-i", "test/felix/io_test/prog.mln-hdfs-combo2,test/felix/io_test/prog.mln-hdfs", "-o", output, 
					"-queryFile", "test/felix/io_test/query.db-hdfs", "-keepData", "-auxSchema",
					"felix_test", "-q", "fact", "-local"};
			
			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
			new Felix().run(options);
			
			String[] expect = { "This is a test.", "This is another sentence.", "This is the end."};
			checkEvidence(expect);
			
			ArrayList<String> lines = FileMan.getLines(output);
			assertTrue(lines.size() == 5);
			for(int i = 0; i < lines.size(); i++) {
				assertTrue(lines.get(i).equals("long(\"This is a test.\")") || 
						lines.get(i).equals("long(\"This is another sentence.\")") ||
						lines.get(i).equals("long(\"This is the end.\")") ||
						lines.get(i).equals("fact(\"11\", \"12\")") || 
						lines.get(i).equals("fact(\"14\", \"15\")"));
			}

		} catch(InterruptedException e) {
			e.printStackTrace();
			assertTrue(false);
		}
	}

	/**
	 * Combo test using evidence generated with Blah Blah, evidence from a file, and evidence from a database table.
	 */
	@Test
	public void comboTest3() {
		try {
			String output = "test/felix/output/out.txt_IOHDFSTest_comboTest3";
			String[] args = {"-e", "test/felix/io_test/evid.db", "-i", 
					"test/felix/io_test/prog.mln-hdfs-combo3,test/felix/io_test/prog.mln-hdfs", "-o", output, 
					"-queryFile", "test/felix/io_test/query.db-hdfs", "-keepData", "-auxSchema",
					"felix_test", "-q", "fact", "-local"};
			
			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
			new Felix().run(options);
			
			String[] expect = { "This is a test.", "This is another sentence.", "This is the end."};
			checkEvidence(expect);
			
			ArrayList<String> lines = FileMan.getLines(output);
			System.out.println("size: " + lines.size());
			assertTrue(lines.size() == 7);
			for(int i = 0; i < lines.size(); i++) {
				System.out.println("line: " + lines.get(i));
				assertTrue(lines.get(i).equals("long(\"This is a test.\")") || 
						lines.get(i).equals("long(\"This is another sentence.\")") ||
						lines.get(i).equals("long(\"This is the end.\")") ||
						lines.get(i).equals("fact(\"1\", \"2\")") || 
						lines.get(i).equals("fact(\"4\", \"5\")") ||
						lines.get(i).equals("fact(\"11\", \"12\")") || 
						lines.get(i).equals("fact(\"14\", \"15\")"));

			}

		} catch(InterruptedException e) {
			e.printStackTrace();
			assertTrue(false);
		}
	}

	/**
	 * Combo test using evidence generated with Blah Blah for a text file and a XML file, evidence from a file, and evidence from a database table.
	 */
	@Test
	public void comboTest4() {
		try {
			String output = "test/felix/output/out.txt_IOHDFSTest_comboTest4";
			String[] args = {"-e", "test/felix/io_test/evid.db", "-i", 
					"test/felix/io_test/prog.mln-hdfs-combo4,test/felix/io_test/prog.mln-hdfs", "-o", output, 
					"-queryFile", "test/felix/io_test/query.db-hdfs", "-keepData", "-auxSchema",
					"felix_test", "-q", "fact,long2", "-local"};
			
			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
			new Felix().run(options);
			
			String[] expect = { "This is a test.", "This is another sentence.", "This is the end."};
			checkEvidence(expect);
			
			ArrayList<String> lines = FileMan.getLines(output);
			System.out.println("size: " + lines.size());
			assertTrue(lines.size() == 10);
			for(int i = 0; i < lines.size(); i++) {
				System.out.println("line: " + lines.get(i));
				assertTrue(lines.get(i).equals("long(\"This is a test.\")") || 
						lines.get(i).equals("long(\"This is another sentence.\")") ||
						lines.get(i).equals("long(\"This is the end.\")") ||
						lines.get(i).equals("long2(\"This is a test.\")") || 
						lines.get(i).equals("long2(\"This is another sentence.\")") ||
						lines.get(i).equals("long2(\"This is the end.\")") ||
						lines.get(i).equals("fact(\"1\", \"2\")") || 
						lines.get(i).equals("fact(\"4\", \"5\")") ||
						lines.get(i).equals("fact(\"11\", \"12\")") || 
						lines.get(i).equals("fact(\"14\", \"15\")"));

			}

		} catch(InterruptedException e) {
			e.printStackTrace();
			assertTrue(false);
		}
	}

	/**
	 * Test re-running with a feature that is already extracted.
	 */
	@Test
	public void repeatTest() {
		try {
			String output = "test/felix/output/out.txt_IOHDFSTest_repeatTest";
			String[] args = {"-i", "test/felix/io_test/prog.mln-hdfs-simple,test/felix/io_test/prog.mln-hdfs", "-o", output, 
					"-queryFile", "test/felix/io_test/query.db-hdfs", "-keepData", "-auxSchema",
					"felix_test", "-local"};
			
			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
			new Felix().run(options);
			
			String[] expect = { "This is a test.", "This is another sentence.", "This is the end." };
			checkEvidence(expect);
			checkOutput(output, expect);
			
			// TODO: check that feature is not re-extracted somehow?

		} catch(InterruptedException e) {
			e.printStackTrace();
			assertTrue(false);
		}
	}

// NOTE: bad test, XML is case sensitive
//	@Test
//	public void xmlCaseTest() {
//		try {
//			String output = "test/felix/output/out.txt_IOHDFSTest_xmlCaseTest";
//			String[] args = {"-i", "test/felix/io_test/prog.mln-hdfs-xml-case,test/felix/io_test/prog.mln-hdfs", "-o", output, 
//					"-queryFile", "test/felix/io_test/query.db-hdfs", "-keepData", "-auxSchema",
//					"felix_test", "-local"};
//			
//			FelixConfig.overrideID();
//			FelixCommandOptions options = FelixUIMan.parseCommand(args);
//			new Felix().run(options);
//			
//			String[] expect = { "This is a test.", "This is another sentence.", "This is the end." };
//			checkEvidence(expect);
//			checkOutput(output, expect);
//		} catch(InterruptedException e) {
//			e.printStackTrace();
//			assertTrue(false);
//		}
//	}
	
	/**
	 * Test using an XML file with no matching tags.
	 */
	@Test
	public void xmlNoMatchTest() {
		try {
			String output = "test/felix/output/out.txt_IOHDFSTest_xmlNoMatchTest";
			String[] args = {"-i", "test/felix/io_test/prog.mln-hdfs-xml-no-match,test/felix/io_test/prog.mln-hdfs", "-o", output, 
					"-queryFile", "test/felix/io_test/query.db-hdfs", "-keepData", "-auxSchema",
					"felix_test", "-local"};
			
			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
			new Felix().run(options);
			
			String[] expect = { };
			checkEvidence(expect);
			checkOutput(output, expect);
		} catch(InterruptedException e) {
			e.printStackTrace();
			assertTrue(false);
		}
	}

	/**
	 * Test the use of python comments.
	 */
	@Test
	public void commentTest() {
		try {
			String output = "test/felix/output/out.txt_IOHDFSTest_commentTest";
			String[] args = {"-i", "test/felix/io_test/prog.mln-hdfs-comment,test/felix/io_test/prog.mln-hdfs", "-o", output, 
					"-queryFile", "test/felix/io_test/query.db-hdfs", "-keepData", "-auxSchema",
					"felix_test", "-local"};
			
			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
			new Felix().run(options);
			
			String[] expect = { "This is a test.", "This is another sentence.", "This is the end."};
			checkEvidence(expect);
			checkOutput(output, expect);

		} catch(InterruptedException e) {
			e.printStackTrace();
			assertTrue(false);
		}
	}
	
	/**
	 * Test the use of python print statements.
	 */
	@Test
	public void printTest() {
		try {
			String output = "test/felix/output/out.txt_IOHDFSTest_printTest";
			String[] args = {"-i", "test/felix/io_test/prog.mln-hdfs-print,test/felix/io_test/prog.mln-hdfs", "-o", output, 
					"-queryFile", "test/felix/io_test/query.db-hdfs", "-keepData", "-auxSchema",
					"felix_test", "-local"};
			
			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
			new Felix().run(options);
			
			String[] expect = { "This is a test.", "This is another sentence.", "This is the end."};
			checkEvidence(expect);
			checkOutput(output, expect);
			
			// TODO: check if print statements actually printed to console?

		} catch(InterruptedException e) {
			e.printStackTrace();
			assertTrue(false);
		}
	}


}
