package felix.test;

import static org.junit.Assert.*;

import java.io.BufferedReader;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.junit.BeforeClass;
import org.junit.Test;

import tuffy.util.UIMan;

import felix.main.Felix;
import felix.parser.FelixCommandOptions;
import felix.util.FelixConfig;
import felix.util.FelixUIMan;


/**
 * Test LR operator's quality.
 * @author Ce Zhang
 *
 */
public class LRTest{
	
	public static ArrayList<Double> planCosts = new ArrayList<Double>();
	public static String exeTime = "";
	

	/**
	 * Test Marginal inference results of LR on a test case with manually-built ground truth.
	 */
	@Test
	public final void testLRMarginal() {
		try{
			String[] args = {"-e", "test/smoke/evidence.db", "-i", "test/smoke/prog.mln",
					 "-o", "test/testOutput.txt", "-queryFile", "test/smoke/query.db", "-marginal"};
			
			for(int wwww = 0; wwww < 2; wwww++){
			
				FelixConfig.overrideID();
				FelixCommandOptions options = FelixUIMan.parseCommand(args);
				
				if(wwww == 1){
					options.useDualDecomposition = true;
					FelixConfig.nDDIT = 2;
				}
				
				new Felix().run(options);
				
				String outfile = "test/testOutput.txt";
				
				if(wwww == 1){
					outfile = outfile + "_snap_0";
				}
				
				//TODO:
				
				//Ground truth:
				//Gray Smoke 0 => 0.59
				//Anna and Edward Smoke 1 => 0.52
				
				/*
				0.52497918747894001	Smokes("Edward", "1")
				0.47502081252106004	Smokes("Edward", "0")
				0.401312339887548	Smokes("Gary", "1")
				0.598687660112452	Smokes("Gary", "0")
				0.52497918747894001	Smokes("Anna", "1")
				0.47502081252106004	Smokes("Anna", "0")
				 */
				
				BufferedReader br = new BufferedReader(new FileReader("test/testOutput.txt"));
				
				Pattern p1 = Pattern.compile("0.52.*Smokes.*Edward.*1");
				Pattern p2 = Pattern.compile("0.47.*Smokes.*Edward.*0");		
				Pattern p3 = Pattern.compile("0.40.*Smokes.*Gary.*1");
				Pattern p4 = Pattern.compile("0.59.*Smokes.*Gary.*0");	
				Pattern p5 = Pattern.compile("0.52.*Smokes.*Anna.*1");
				Pattern p6 = Pattern.compile("0.47.*Smokes.*Anna.*0");
				
				String line;
				while( (line = br.readLine()) != null){
					int ct = 0;
					
					Matcher m = p1.matcher(line);
					if(m.find()){
						ct++;
					}
					
					m = p2.matcher(line);
					if(m.find()){
						ct++;
					}
					
					m = p3.matcher(line);
					if(m.find()){
						ct++;
					}
					
					m = p4.matcher(line);
					if(m.find()){
						ct++;
					}
					
					m = p5.matcher(line);
					if(m.find()){
						ct++;
					}
					
					m = p6.matcher(line);
					if(m.find()){
						ct++;
					}
					
					UIMan.warn(""+ct);
					assertTrue(ct == 1);
					
					
				}
			}
			
		}catch(Exception e){
			e.printStackTrace();
		}
	}
	
	/**
	 * Test MAP inference results of LR on a test case with manually-built ground truth.
	 */
	@Test
	public final void testLRMAP() {
		try{
			String[] args = {"-e", "test/smoke/evidence.db", "-i", "test/smoke/prog.mln",
					"-o", "test/testOutput.txt", "-queryFile", "test/smoke/query.db"};
			
			
			for(int wwww = 0; wwww < 2; wwww++){
				
				FelixConfig.overrideID();
				FelixCommandOptions options = FelixUIMan.parseCommand(args);
				
				if(wwww == 1){
					options.useDualDecomposition = true;
					FelixConfig.nDDIT = 2;
				}
			
				String outfile = "test/testOutput.txt";
				
				if(wwww == 1){
					outfile = outfile + "_snap_1";
				}
			
				new Felix().run(options);
				
				//Ground truth:
				//Gray Smoke 0
				//Anna and Edward Smoke 1
	
				/*
				Smokes("Edward", "1")
				Smokes("Anna", "1")
				Smokes("Gary", "0")
				*/
				
				BufferedReader br = new BufferedReader(new FileReader(outfile));
				
				Pattern p1 = Pattern.compile("Smokes.*Edward.*1");		
				Pattern p4 = Pattern.compile("Smokes.*Gary.*0");	
				Pattern p5 = Pattern.compile("Smokes.*Anna.*1");
				
				String line;
				while( (line = br.readLine()) != null){
					int ct = 0;
					
					Matcher m = p1.matcher(line);
					if(m.find()){
						ct++;
					}
					
					m = p4.matcher(line);
					if(m.find()){
						ct++;
					}
					
					m = p5.matcher(line);
					if(m.find()){
						ct++;
					}
	
					assertTrue(ct == 1);
					
				}

			}
			
		}catch(Exception e){
			e.printStackTrace();
			assertTrue(false);
		}
	}
	
	
	
}






