package felix.test;

import static org.junit.Assert.*;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;


import org.junit.BeforeClass;
import org.junit.Test;

import tuffy.util.Timer;
import tuffy.util.UIMan;
import felix.compiler.StaticAnalyzer;
import felix.dstruct.ConcurrentOperatorsBucket;
import felix.dstruct.ExecutionPlan;
import felix.dstruct.StatOperator;
import felix.executor.Executor;
import felix.main.Felix;
import felix.optimizer.Scheduler;
import felix.parser.FelixCommandOptions;
import felix.util.FelixConfig;
import felix.util.FelixUIMan;


/**
 * Test the scheduler: 1) cost-based optimizer; 2) planner of the
 * operator order.
 * @author Ce Zhang
 *
 */
public class SchedulerTest extends Felix{
	
	public static ArrayList<Double> planCosts = new ArrayList<Double>();
	public static String exeTime = "";
	
	/**
	 * Test the planner for operator order: One operator A should run before
	 * operator B if A provide information to B or A has a larger precedence.
	 */
	@Test
	public final void testSillyPlanner() {
		try{
			String[] args = {"-e", "test/felix/compiler_test_evidence.db", "-i", "test/felix/compiler_test_prog.mln",
					"-o", "test/testOutput.txt", "-queryFile", "test/felix/compiler_test_query.db"};
			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);


			Timer.start("Felix-Timer");
			this.options = options;
			resetACoupleAuxDataStructures();

			fq = this.parseFelixQuery();

			sa = new StaticAnalyzer(this.fq, options);
			sa.parse();

			Scheduler sc = new Scheduler(this, this.fq, options);
			ExecutionPlan ep = sc.schedule();
			
			for(int i = 0; i < ep.operators.size(); i++){
				for(int j = 0; j < ep.operators.size(); j ++){
					
					if( i >= j ){
						continue;
					}
					
					ConcurrentOperatorsBucket oup = ep.operators.get(j);
					ConcurrentOperatorsBucket odown = ep.operators.get(i);
					
					if(i < j){
						if(sc.getOperatorBucketGraph().getDownStreamOperator(oup).contains(odown) && 
								sc.getOperatorBucketGraph().getDownStreamOperator(odown).contains(oup)){
							assertTrue(oup.getPrecedence() >= odown.getPrecedence());
						}else if(sc.getOperatorBucketGraph().getDownStreamOperator(oup).contains(odown)){
							assertTrue(true);
						}else if(sc.getOperatorBucketGraph().getDownStreamOperator(odown).contains(oup)){
							assertTrue(false);
						}else{
							assertTrue(true);
						}
					}
					
				}
			}	
			
		}catch(Exception e){
			e.printStackTrace();
		}
	}
	
	/**
	 * Test the cost-based optimizer on a sampled Enron dataset.
	 */
	@Test
	public final void testDecomposition(){
		
		try{
/*
			double minCost = Double.MAX_VALUE;
			double nTime = -1;
			
			int numberOfIters = 10;
			String tobprint = "";
			double sum = 0;
			for(int i=0;i<numberOfIters;i++){
				sum = 0;
				Config.pickRandom = true;
				String[] args = {"-e", "test/mobius/enron.evid.db.5k", "-i", "test/mobius/enron.prog.mln--",
						"-o", "test/testOutput.txt", "-queryFile", "test/mobius/query.db", "-mcsatSamples", "0"};
				
				CommandOptions opt = UIMan.parseCommand(args);
				Config.explainMode = opt.explainMode;
				planCosts.clear();
				
				new Mobius().run(opt);
				
				String time = exeTime;
				for(Double d : planCosts){
					sum += d;
				}
				Config.pickRandom = false;
				
				tobprint += "\nEstimated Cost : " + sum + "\t|\t Execution Time: " + time + "\n";
				tobprint += planCosts.toString() + "\n";
				System.err.println(tobprint);
				
				//Manually check whether output is reasonable
				
			}
			*/
			
		}catch(Exception e){
			e.printStackTrace();
			assertTrue(false);
		}

	}
	
}






