package felix.test;

import static org.junit.Assert.*;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;


import org.junit.BeforeClass;
import org.junit.Test;

import tuffy.util.Config;
import tuffy.util.Timer;
import tuffy.util.UIMan;

import felix.compiler.StaticAnalyzer;
import felix.dstruct.ConcurrentOperatorsBucket;
import felix.dstruct.ExecutionPlan;
import felix.dstruct.FelixPredicate;
import felix.dstruct.FelixPredicate.FPProperty;
import felix.dstruct.StatOperator.OPType;
import felix.dstruct.OperatorBucketGraph;
import felix.dstruct.StatOperator;
import felix.executor.Executor;
import felix.main.Felix;
import felix.optimizer.Scheduler;
import felix.parser.FelixCommandOptions;
import felix.util.FelixConfig;
import felix.util.FelixUIMan;

/**
 * Test the compiler.
 * @author Ce Zhang
 *
 */
public class StaticAnalyzerTest extends Felix{
	
	//this.parseKeyConstraintRelation();
	//this.parseReflexiveRelation();
	//this.parseSymmetricRelation();
	//this.parseTransitiveRelation();
	//this.parseChainRecursiveRelation();
	//this.parseNonRecursiveRelation();
	//this.parseOtherRecursiveRelation();
	//this.parseSpecialPredicate();
	//this.parseEmbededWeightRule();
	
	/**
	 * Test the parser for CRF operator.
	 */
	@Test
	public final void test_parseKeyConstraintRelation() {
		try{
			String[] args = {"-e", "test/felix/compiler_test_evidence.db", "-i", "test/felix/compiler_test_prog.mln",
					"-o", "test/testOutput.txt", "-queryFile", "test/felix/compiler_test_query.db"};

			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
				
			Timer.start("Felix-Timer");
			this.options = options;
			resetACoupleAuxDataStructures();

			fq = this.parseFelixQuery();

			sa = new StaticAnalyzer(this.fq, options);
			
			sa.parseKeyConstraintRelation();
			
			int ct = 0;
			for(FelixPredicate fp : this.fq.getAllPred()){
				
				if(fp.getName().equals("lrPred1")){
					ct++;
					assertTrue(fp.hasProperty(FPProperty.KEY_CONSTRAINT));
					assertTrue(fp.getLabelPositions().size() == 1);
					assertTrue(fp.getKeyPositions().size() == 5);
					assertTrue(fp.getLabelFieldsArgs().size() == 1);
					assertTrue(fp.getKeyFieldsArgs().size() == 5);
					
					assertTrue(fp.getLabelPositions().iterator().next() == 4);

				}else if(fp.getName().equals("lrPred2")){
					ct++;
					assertTrue(fp.hasProperty(FPProperty.KEY_CONSTRAINT));
					assertTrue(fp.getLabelPositions().size() == 1);
					assertTrue(fp.getKeyPositions().size() == 5);
					assertTrue(fp.getLabelFieldsArgs().size() == 1);
					assertTrue(fp.getKeyFieldsArgs().size() == 5);
					
					assertTrue(fp.getLabelPositions().iterator().next() == 4);
					
				}else if(fp.getName().equals("crfPred")){
					ct++;
					assertTrue(fp.hasProperty(FPProperty.KEY_CONSTRAINT));
					assertTrue(fp.getLabelPositions().size() == 1);
					assertTrue(fp.getKeyPositions().size() == 5);
					assertTrue(fp.getLabelFieldsArgs().size() == 1);
					assertTrue(fp.getKeyFieldsArgs().size() == 5);
					
					assertTrue(fp.getLabelPositions().iterator().next() == 4);
					
				}else if(fp.getName().equals("notCrfPred")){
					ct++;
					assertTrue(fp.hasProperty(FPProperty.KEY_CONSTRAINT));
					assertTrue(fp.getLabelPositions().size() == 1);
					assertTrue(fp.getKeyPositions().size() == 5);
					assertTrue(fp.getLabelFieldsArgs().size() == 1);
					assertTrue(fp.getKeyFieldsArgs().size() == 5);
					
					assertTrue(fp.getLabelPositions().iterator().next() == 4);
					
				}else{
					assertFalse(fp.hasProperty(FPProperty.KEY_CONSTRAINT));
					assertTrue(fp.getKeyPositions().size() == 2);
					assertTrue(fp.getKeyFieldsArgs().size() == 2);
					assertTrue(fp.getLabelFieldsArgs().size() == 0);
				}
			}
			
			assertTrue(ct == 4);
			
		}catch(Exception e){
			e.printStackTrace();
		}
		
	}
	

	@Test
	public final void test_parseChainRecursiveRelation() {
		try{
			String[] args = {"-e", "test/felix/compiler_test_evidence.db", "-i", "test/felix/compiler_test_prog.mln",
					"-o", "test/testOutput.txt", "-queryFile", "test/felix/compiler_test_query.db"};

			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
				
			Timer.start("Felix-Timer");
			this.options = options;
			resetACoupleAuxDataStructures();

			fq = this.parseFelixQuery();

			sa = new StaticAnalyzer(this.fq, options);
			
			sa.parseKeyConstraintRelation();
			sa.parseChainRecursiveRelation();
			
			int ct = 0;
			for(FelixPredicate fp : this.fq.getAllPred()){
				
				if(fp.getName().equals("crfPred")){
					ct++;
					assertTrue(fp.hasProperty(FPProperty.CHAIN_RECUR));
					assertTrue(fp.getPropertyClauses(FPProperty.CHAIN_RECUR).size() == 1);
					assertTrue(fp.getPropertyClauses(FPProperty.CHAIN_RECUR).iterator().next().getWeight() == -3.14);
					
				}else{
					assertFalse(fp.hasProperty(FPProperty.CHAIN_RECUR));
					assertTrue(fp.getPropertyClauses(FPProperty.CHAIN_RECUR).size() == 0);
				}
			}
			
			assertTrue(ct == 1);
			
		}catch(Exception e){
			e.printStackTrace();
		}
		
	}
	
	@Test
	public final void test_parseEmbededWeightRule() {
		try{
			String[] args = {"-e", "test/felix/compiler_test_evidence.db", "-i", "test/felix/compiler_test_prog.mln",
					"-o", "test/testOutput.txt", "-queryFile", "test/felix/compiler_test_query.db"};

			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
				
			Timer.start("Felix-Timer");
			this.options = options;
			resetACoupleAuxDataStructures();

			fq = this.parseFelixQuery();

			sa = new StaticAnalyzer(this.fq, options);
			
			sa.parseEmbededWeightRule();
			
			int ct = 0;
			for(FelixPredicate fp : this.fq.getAllPred()){
				
				if(fp.getName().equals("lrPred2")){
					ct++;
					assertTrue(fp.hasProperty(FPProperty.EMBED_WEIGHT_RULE));
					assertTrue(fp.getPropertyClauses(FPProperty.EMBED_WEIGHT_RULE).size() == 2);

				}else if(fp.getName().equals("notCorefPred")){
					ct++;
					assertTrue(fp.hasProperty(FPProperty.EMBED_WEIGHT_RULE));
					assertTrue(fp.getPropertyClauses(FPProperty.EMBED_WEIGHT_RULE).size() == 1);

				}else{
					assertFalse(fp.hasProperty(FPProperty.CHAIN_RECUR));
					assertTrue(fp.getPropertyClauses(FPProperty.CHAIN_RECUR).size() == 0);
				}
			}
			
			assertTrue(ct == 2);
			
		}catch(Exception e){
			e.printStackTrace();
		}
		
	}
	
	
	@Test
	public final void test_parseNonRecursiveRelation() {
		try{
			String[] args = {"-e", "test/felix/compiler_test_evidence.db", "-i", "test/felix/compiler_test_prog.mln",
					"-o", "test/testOutput.txt", "-queryFile", "test/felix/compiler_test_query.db"};

			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
				
			Timer.start("Felix-Timer");
			this.options = options;
			resetACoupleAuxDataStructures();

			fq = this.parseFelixQuery();

			sa = new StaticAnalyzer(this.fq, options);
			
			sa.parseNonRecursiveRelation();
			
			int ct = 0;
			for(FelixPredicate fp : this.fq.getAllPred()){
				
				if(fp.getName().equals("crfPred")){
					ct++;
					assertFalse(fp.hasProperty(FPProperty.NON_RECUR));
					assertTrue(fp.getPropertyClauses(FPProperty.NON_RECUR).size() == 0);

				}else if(fp.getName().equals("notCrfPred")){
					ct++;
					assertFalse(fp.hasProperty(FPProperty.NON_RECUR));
					assertTrue(fp.getPropertyClauses(FPProperty.NON_RECUR).size() == 0);

				}else{
					if(fp.isClosedWorld()){
						continue;
					}
					assertTrue(fp.hasProperty(FPProperty.NON_RECUR));
					assertTrue(fp.getPropertyClauses(FPProperty.NON_RECUR).size() > 0);
				}
			}
			
			assertTrue(ct == 2);
			
		}catch(Exception e){
			e.printStackTrace();
		}
		
	}
	
	@Test
	public final void test_parseOtherRecursiveRelation() {
		try{
			String[] args = {"-e", "test/felix/compiler_test_evidence.db", "-i", "test/felix/compiler_test_prog.mln",
					"-o", "test/testOutput.txt", "-queryFile", "test/felix/compiler_test_query.db"};

			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
				
			Timer.start("Felix-Timer");
			this.options = options;
			resetACoupleAuxDataStructures();

			fq = this.parseFelixQuery();

			sa = new StaticAnalyzer(this.fq, options);
			
			sa.parseKeyConstraintRelation();
			sa.parseReflexiveRelation();
			sa.parseSymmetricRelation();
			sa.parseTransitiveRelation();
			sa.parseNonRecursiveRelation();
			sa.parseChainRecursiveRelation();
			sa.parseOtherRecursiveRelation();
			
			int ct = 0;
			for(FelixPredicate fp : this.fq.getAllPred()){
				
				if(fp.getName().equals("notCrfPred")){
					ct++;
					assertTrue(fp.hasProperty(FPProperty.OTHER_RECUR));
					assertTrue(fp.getPropertyClauses(FPProperty.OTHER_RECUR).size() == 1);

				}else{
					System.out.println(fp.getName());
					assertFalse(fp.hasProperty(FPProperty.OTHER_RECUR));
					assertTrue(fp.getPropertyClauses(FPProperty.OTHER_RECUR).size() == 0);
				}
			}
			
			assertTrue(ct == 1);
			
		}catch(Exception e){
			e.printStackTrace();
		}
		
	}
	
	@Test
	public final void test_parseReflexiveRelation() {
		try{
			String[] args = {"-e", "test/felix/compiler_test_evidence.db", "-i", "test/felix/compiler_test_prog.mln",
					"-o", "test/testOutput.txt", "-queryFile", "test/felix/compiler_test_query.db"};

			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
				
			Timer.start("Felix-Timer");
			this.options = options;
			resetACoupleAuxDataStructures();

			fq = this.parseFelixQuery();

			sa = new StaticAnalyzer(this.fq, options);
			
			sa.parseKeyConstraintRelation();
			sa.parseReflexiveRelation();
			
			int ct = 0;
			for(FelixPredicate fp : this.fq.getAllPred()){
				
				if(fp.getName().equals("corefPred1")){
					ct++;
					assertTrue(fp.hasProperty(FPProperty.REFLEX));
					assertTrue(fp.getPropertyClauses(FPProperty.REFLEX).size() == 1);

				}else if(fp.getName().equals("corefPred2")){
					ct++;
					assertTrue(fp.hasProperty(FPProperty.REFLEX));
					assertTrue(fp.getPropertyClauses(FPProperty.REFLEX).size() == 1);

				}else if(fp.getName().equals("notCorefPred")){
					ct++;
					assertTrue(fp.hasProperty(FPProperty.REFLEX));
					assertTrue(fp.getPropertyClauses(FPProperty.REFLEX).size() == 1);

				}else{
					assertFalse(fp.hasProperty(FPProperty.REFLEX));
					assertTrue(fp.getPropertyClauses(FPProperty.REFLEX).size() == 0);
				}
			}
			
			assertTrue(ct == 3);
			
		}catch(Exception e){
			e.printStackTrace();
		}
		
	}
	
	@Test
	public final void test_parseSymmetricRelation() {
		try{
			String[] args = {"-e", "test/felix/compiler_test_evidence.db", "-i", "test/felix/compiler_test_prog.mln",
					"-o", "test/testOutput.txt", "-queryFile", "test/felix/compiler_test_query.db"};

			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
				
			Timer.start("Felix-Timer");
			this.options = options;
			resetACoupleAuxDataStructures();

			fq = this.parseFelixQuery();

			sa = new StaticAnalyzer(this.fq, options);
			
			sa.parseKeyConstraintRelation();
			sa.parseSymmetricRelation();
			
			int ct = 0;
			for(FelixPredicate fp : this.fq.getAllPred()){
				
				if(fp.getName().equals("corefPred1")){
					ct++;
					assertTrue(fp.hasProperty(FPProperty.SYMM));
					assertTrue(fp.getPropertyClauses(FPProperty.SYMM).size() == 1);

				}else if(fp.getName().equals("corefPred2")){
					ct++;
					assertTrue(fp.hasProperty(FPProperty.SYMM));
					assertTrue(fp.getPropertyClauses(FPProperty.SYMM).size() == 2);

				}else if(fp.getName().equals("notCorefPred")){
					ct++;
					assertTrue(fp.hasProperty(FPProperty.SYMM));
					assertTrue(fp.getPropertyClauses(FPProperty.SYMM).size() == 1);

				}else{
					assertFalse(fp.hasProperty(FPProperty.SYMM));
					assertTrue(fp.getPropertyClauses(FPProperty.SYMM).size() == 0);
				}
			}
			
			assertTrue(ct == 3);
			
		}catch(Exception e){
			e.printStackTrace();
		}
		
	}
	
	@Test
	public final void test_parseTransitiveRelation() {
		try{
			String[] args = {"-e", "test/felix/compiler_test_evidence.db", "-i", "test/felix/compiler_test_prog.mln",
					"-o", "test/testOutput.txt", "-queryFile", "test/felix/compiler_test_query.db"};

			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
				
			Timer.start("Felix-Timer");
			this.options = options;
			resetACoupleAuxDataStructures();

			fq = this.parseFelixQuery();

			sa = new StaticAnalyzer(this.fq, options);
			
			sa.parseKeyConstraintRelation();
			sa.parseTransitiveRelation();
			
			int ct = 0;
			for(FelixPredicate fp : this.fq.getAllPred()){
				
				if(fp.getName().equals("corefPred1")){
					ct++;
					assertTrue(fp.hasProperty(FPProperty.TRANS));
					assertTrue(fp.getPropertyClauses(FPProperty.TRANS).size() == 1);

				}else if(fp.getName().equals("corefPred2")){
					ct++;
					assertTrue(fp.hasProperty(FPProperty.TRANS));
					assertTrue(fp.getPropertyClauses(FPProperty.TRANS).size() == 1);

				}else if(fp.getName().equals("notCorefPred")){
					ct++;
					assertTrue(fp.hasProperty(FPProperty.TRANS));
					assertTrue(fp.getPropertyClauses(FPProperty.TRANS).size() == 1);

				}else{
					assertFalse(fp.hasProperty(FPProperty.TRANS));
					assertTrue(fp.getPropertyClauses(FPProperty.TRANS).size() == 0);
				}
			}
			
			assertTrue(ct == 3);
			
		}catch(Exception e){
			e.printStackTrace();
		}
		
	}
	
	@Test
	public final void test_parseSpecialPredicate() {
		try{
			String[] args = {"-e", "test/felix/compiler_test_evidence.db", "-i", "test/felix/compiler_test_prog.mln",
					"-o", "test/testOutput.txt", "-queryFile", "test/felix/compiler_test_query.db"};

			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
				
			Timer.start("Felix-Timer");
			this.options = options;
			resetACoupleAuxDataStructures();

			fq = this.parseFelixQuery();

			sa = new StaticAnalyzer(this.fq, options);
			
			sa.parseKeyConstraintRelation();
			sa.parseReflexiveRelation();
			sa.parseSymmetricRelation();
			sa.parseTransitiveRelation();
			sa.parseChainRecursiveRelation();
			sa.parseNonRecursiveRelation();
			sa.parseOtherRecursiveRelation();
			sa.parseSpecialPredicate();
			
			int ct = 0;
			for(FelixPredicate fp : this.fq.getAllPred()){
				
				if(fp.getName().equals("corefPred1_map")){
					ct++;
					assertTrue(fp.isCorefMapPredicate == true);
				}else{
					assertFalse(fp.isCorefMapPredicate == true);
				}
			}
			
			assertTrue(ct == 1);
			
		}catch(Exception e){
			e.printStackTrace();
		}
		
	}
	
	
	/**
	 * Run Felix on a very simple program.
	 */
	@Test
	public final void naiveRun(){
		try{
			String[] args = {"-e", "test/felix/compiler_test_evidence.db", "-i", "test/felix/compiler_test_prog.mln",
					"-o", "test/testOutput.txt", "-queryFile", "test/felix/compiler_test_query.db", "-threads", "1"};
			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);
			new Felix().run(options);
			
			String[] args1 = {"-e", "test/felix/compiler_test_evidence.db", "-i", "test/felix/compiler_test_prog.mln",
					"-o", "test/testOutput.txt", "-queryFile", "test/felix/compiler_test_query.db", "-marginal", "-threads", "1"};
			FelixConfig.overrideID();
			options = FelixUIMan.parseCommand(args);
			new Felix().run(options);
						
		}catch(Exception e){
			e.printStackTrace();
			assertTrue(false);
		}
	}
	
	/**
	 * Test whether the parser and og graph can deal with dependency correctly.
	 */
	@Test
	public final void testParse() {
		try{
			String[] args = {"-e", "test/felix/compiler_test_evidence.db", "-i", "test/felix/compiler_test_prog.mln",
					"-o", "test/testOutput.txt", "-queryFile", "test/felix/compiler_test_query.db"};
			FelixConfig.overrideID();
			FelixCommandOptions options = FelixUIMan.parseCommand(args);


			Timer.start("Felix-Timer");
			this.options = options;
			resetACoupleAuxDataStructures();

			fq = this.parseFelixQuery();

			sa = new StaticAnalyzer(this.fq, options);
			sa.parse();

			Scheduler sc = new Scheduler(this, this.fq, options);
			ExecutionPlan ep = sc.schedule();
			
			int ncrf = 0;
			int nlr = 0;
			int ncoref = 0;
			int ntuffy = 0;
			OperatorBucketGraph og = sc.getOperatorBucketGraph();
			
			for(ConcurrentOperatorsBucket cob : sc.getOperatorBucketGraph().getOperators()){
				for(StatOperator sop : cob.getOperators()){
					if(sop.type == OPType.TUFFY){
						assertTrue(og.getDownStreamOperator(cob).size() == 0);
						assertTrue(og.getUpStreamOperator(cob).size() == 2);
						//assertTrue(og.getDownStreamOperator(cob).iterator().next().type == OPType.COREF);
						assertTrue(og.getUpStreamOperator(cob).iterator().next().type == OPType.COREF);
						//assertTrue(sop.folClauses.size() == 7);
						ntuffy ++;
					}
					if(sop.type == OPType.COREF){
						assertTrue(og.getDownStreamOperator(cob).size() == 1 || og.getDownStreamOperator(cob).size() == 0);
						assertTrue(og.getUpStreamOperator(cob).size() == 1 || og.getUpStreamOperator(cob).size() == 0);
						
						if(og.getUpStreamOperator(cob).size() == 1){
							assertTrue(og.getDownStreamOperator(cob).iterator().next().type == OPType.TUFFY);
							assertTrue(og.getUpStreamOperator(cob).iterator().next().type == OPType.TUFFY);
						}
						
						ncoref ++;
					}
					if(sop.type == OPType.LR){
						assertTrue(og.getDownStreamOperator(cob).size() == 0 && og.getDownStreamOperator(cob).size() == 0);
						nlr ++;
					}
					if(sop.type == OPType.CRF){
						assertTrue(og.getDownStreamOperator(cob).size() == 0 && og.getDownStreamOperator(cob).size() == 0);
						ncrf ++;
					}
				}
			}
			
			assertTrue(nlr == 2 * (Config.getNumThreads()-1));
			assertTrue(ncrf == 1 * (Config.getNumThreads()-1));
			assertTrue(ncoref == 3);
			assertTrue(ntuffy == 1);
			
			
		}catch(Exception e){
			e.printStackTrace();
		}
	}
	
}






