package felix.util;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.math.BigInteger;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;

import tuffy.db.RDB;
import tuffy.mln.Literal;
import tuffy.mln.Predicate;
import tuffy.mln.Term;
import tuffy.util.Config;
import tuffy.util.ExceptionMan;
import tuffy.util.FileMan;
import tuffy.util.StringMan;
import tuffy.util.UIMan;
import felix.dstruct.FelixClause;
import felix.dstruct.FelixPredicate;
import felix.dstruct.FelixQuery;
import felix.parser.FelixCommandOptions;

public class ActiveLearner {

	FelixCommandOptions opt;

	RDB db;

	// TODO: make this general
	/**
	 * Generates xml file containing relevant info for query.
	 * @throws Exception
	 */
	public void getTopCandidates() throws Exception{

		FelixUIMan.println(">>> Connecting to RDBMS at " + FelixConfig.db_url);

		if(FelixConfig.evidDBSchema == null){
			ExceptionMan.die("Active learning mode needs the evidence schema specified");
		}

		if(FelixConfig.evidDBSchema != null){
			FelixConfig.db_schema = FelixConfig.evidDBSchema;
		}

		db = RDB.getRDBbyConfig();
		db.resetSchema(FelixConfig.db_schema);

		RDB db2 = RDB.getRDBbyConfig();
		db2.resetSchema(FelixConfig.db_schema);

		HashMap<Long, String> id2str = db.loadIdSymbolMapFromTable();
		HashMap<String, Integer> str2id = db.loadSymbolIdMapFromTable();

		FelixQuery fq = new FelixQuery();

		String[] progFiles = opt.fprog.split(",");
		fq.loadPrograms(progFiles);

		if(opt.fquery != null){
			String[] queryFiles = opt.fquery.split(",");
			fq.loadQueries(queryFiles);
		}

		if(opt.queryAtoms != null){
			FelixUIMan.println(">>> Parsing query atoms in command line");
			fq.parseQueryCommaList(opt.queryAtoms);
		}

		// Predicate -> (attr -> value)
		/*
		 predicate |  property  |    value
		-----------+------------+-------------
		 el        | optype     | lr
		 el        | provtable  | pred_DMO_43
		 el        | labelfield | eid2
		 */
		HashMap<FelixPredicate, HashMap<String, String>> attrs = new 
				HashMap<FelixPredicate, HashMap<String, String>>();

		// TODO: use this
		//String[] viewDesps = {"QueryWord(queryword) := el(query, entityid), query( query, queryWord, docID)",
		//"EntityWord(wikiTitle, entityWord) := el(query, entityid), entity( entityid, wikiTitle, entityWord)"};

		// TODO: use this!
		//		db.execute("CREATE TEMPORARY VIEW _active_view1 AS SELECT " +
		//				" t0.queryid1 AS queryid1, t0.entityid2 as entityid2, ARRAY[t1.queryword2] as viewcontent  "+
		//				" FROM pred_el t0, pred_query t1 WHERE t0.queryid1 = t1.queryid1");
		//
		//		db.execute("CREATE TEMPORARY VIEW _active_view2 AS SELECT " +
		//				" t0.queryid1 AS queryid1, t0.entityid2 as entityid2, ARRAY[t1.wikititle2, t1.entityword3] as viewcontent "+
		//				" FROM pred_el t0, pred_entity t1 WHERE t0.entityid2 = t1.entityid1");

		// load meta information
		ResultSet rs = db.query("SELECT * from _prov_metainfo;");
		while(rs.next()){
			String pred = rs.getString(1);
			String attr = rs.getString(2);
			String value = rs.getString(3);

			FelixPredicate p = fq.getPredByName(pred);

			if(!attrs.containsKey(p)){
				attrs.put(p, new HashMap<String, String>());
			}
			attrs.get(p).put(attr, value);
		}

		BufferedWriter bw = FileMan.getBufferedWriterMaybeGZ(opt.fout);

		bw.append("<?xml version=\"1.0\"?>\n");
		bw.append("<root>\n\n");


		// TODO: make this work for predicates with arity != 2
		HashMap<Integer, Integer> gt = new HashMap<Integer, Integer>();

		String predicate_name = "";

		// parse gt file
//		if(opt.gtFile != null) {
//			System.out.println("Parsing gt file " + opt.gtFile);
//			try {
//				BufferedReader in = new BufferedReader(new FileReader(opt.gtFile));
//				String str;
//				while ((str = in.readLine()) != null) {
//					predicate_name = str.substring(0, str.indexOf('('));
//					String rest = str.substring(str.indexOf('(')+1, str.lastIndexOf(')') - str.indexOf('(') + 1);
//					String[] args = rest.split(",");
//
//					// Remove leading/trailing quotes
//					for(int i = 0; i < args.length; i++) {
//						args[i] = args[i].trim();
//						if(args[i].startsWith("\"")) {
//							args[i] = args[i].substring(1);
//						}
//						if(args[i].endsWith("\"")) {
//							args[i] = args[i].substring(0, args[i].length() - 1);
//						}
//					}
//					//System.out.println(args[0] + "\t" + args[1]);
//					//System.out.println("\t" + str2id.get(args[0]) + "\t" + str2id.get(args[1]));
//					gt.put(str2id.get(args[0]), str2id.get(args[1]));
//				}
//				in.close();
//			} catch (IOException e) {
//				ExceptionMan.die("Error reading from file " + opt.gtFile);
//			}
//		}

		// Go through each gt predicate
		//for(int h = 0; h < gt.size(); h++) {



		FelixPredicate p = fq.getPredByName(predicate_name);
		ArrayList<String> arg_names = p.getArgs();

		//for(FelixPredicate p : fq.getAllPred()){
		if(p.hasQuery()){

			if(!attrs.get(p).get("optype").equals("lr")){
				ExceptionMan.die("Currently, I can only deal with LR!");
			}

			// this contains: eid2
			ArrayList<String> labelfields = 
					new ArrayList<String>(Arrays.asList(attrs.get(p).get("labelfield").split(",")));

			// this works for ouput of every gt link
			//				String where = arg_names.get(0) +y "=" + str2id.get(gt_args[0]);
			//				for(int i = 1; i < arg_names.size(); i++) {
			//					where += " and " + arg_names.get(i) + "=" + str2id.get(gt_args[i]);
			//				}
			//				rs = db.query("select * from " + p.getRelName() + " where " + where);

			//String select = "id, truth, max(prior), club, atomid";
			//for(int i = 1; i < arg_names.size(); i++) {
			//	select += "," + arg_names.get(i);
			//}
			
			//ArrayList<Integer> temp = new ArrayList<Integer>(gt.keySet());
			//ArrayList<String> where = new ArrayList<String>();
			//for(int i = 0; i < temp.size(); i++) {
			//	where.add(arg_names.get(0) + "=" + temp.get(i));
			//}
			
			// TODO: make this general (check key constraint?)
			// TODO: make this faster (use GROUP BY?)
			// TODO: use JOIN (create table for gt)
			//String query = "SELECT * FROM " + p.getRelName() + " WHERE " + StringMan.join(" OR ", where) + " ORDER BY " + arg_names.get(0) + " ASC, prior DESC";
			String query = "SELECT * FROM " + p.getRelName() + " ORDER BY " + arg_names.get(0) + " ASC, prior DESC";
			System.out.println("rs query: " + query);
			rs = db.query(query);

			/*
 id | truth |        prior         | club | atomid | mid1 |  eid2
----+-------+----------------------+------+--------+------+---------
  1 | t     |                    1 |    2 |        |    5 | 1203067
  2 | t     | 3.72007597602084e-44 |    2 |        |    5 |       1
  3 | t     | 2.67863696180808e-33 |    2 |        |    5 | 1199200
  4 | t     | 6.71418428821159e-79 |    2 |        |    9 |  909864
  5 | t     | 2.05388455404083e-85 |    2 |        |    9 |       1
  6 | t     | 6.71418428821159e-79 |    2 |        |    9 | 1744854
  7 | t     | 6.71418428821159e-79 |    2 |        |    9 | 1300787
  8 | t     | 6.71418428821159e-79 |    2 |        |    9 |  818955
  9 | t     | 1.80485138784542e-35 |    2 |        |    9 | 1590722
 10 | t     | 6.71418428821159e-79 |    2 |        |    9 |  299086
			 */

			int cur_id = 0;
			int choice = 0;
			int output_count = 0;
			boolean found = false;

			while(rs.next() && output_count < 10) {

				// Mention id (felix id)
				Integer id1 = rs.getInt(p.getArgs().get(0));
				
				// Entity id (felix id)
				Integer id2 = rs.getInt(p.getArgs().get(1));

				// check if there's a gt for this result
				if(gt.get(id1) == null) {
					continue;
				}
				
				// Mention id
				String str1 = id2str.get(id1);
				
				// Entity id
				String str2 = id2str.get(id2);
				
				// New id
				if(id1 != cur_id) {
					cur_id = id1;
					choice = 0;
					found = false;
				}
				
				// check gt
				boolean match = (id2.equals(gt.get(id1)));
				
				// Already found correct result
				if(found) {
					continue;
				}
				// result found
				else if(match) {
					found = true;
					// correct
					if(choice == 0) {
						System.out.println(str1 + "\t" + str2);
						System.out.println("\tCorrect");
						continue;
					}
					// otherwise output result
				}
				// incorrect
				else {
					if(choice != 0) {
						choice++;
						continue;
					}
					choice++;
					System.out.println(str1 + "\t" + str2);
					System.out.println("\tIncorrect\t" + id2str.get(gt.get(id1)));
				}
				
				output_count++;				

				//String mmmmmid = id2str.get(rs.getInt(p.getArgs().get(0)));
				//System.out.println(mmmmmid);

				ArrayList<String> binds = new ArrayList<String>();
				ArrayList<String> joincond = new ArrayList<String>();
				ArrayList<String> smalllabelfields = new ArrayList<String>();

				// Go through each column
				for(int i=0;i<p.arity();i++){
					if(!labelfields.contains(p.getArgs().get(i))){
						joincond.add("t0." + p.getArgs().get(i) + "=" + rs.getString(p.getArgs().get(i)));
					}else{
						smalllabelfields.add(id2str.get(rs.getInt(p.getArgs().get(i))));
					}
					binds.add("t0." + p.getArgs().get(i) + "=" + rs.getInt(p.getArgs().get(i)));
				}

				String label = StringMan.commaList(smalllabelfields);
				Double prob = rs.getDouble("prior");
				String probability = UIMan.decimalRound(4, prob);

				//TODO: THINK ABOUT ``NO-KEY'' CASE
				// SELECT * FROM pred_DMO_39 WHERE t0.mid1=163545
				// TODO: only do this once per mention
				String sql = "SELECT * FROM " + attrs.get(p).get("provtable") + " t0 WHERE " +
						StringMan.join(" AND ",joincond) + " ORDER BY t0.weight DESC";
				System.out.println("prov query: " + sql);

				ResultSet provrs = db2.query(sql);
				//					System.out.print(p.getName() + "(");
				bw.append("<QUERY>\n");
				bw.append("  <NAME>" + p.getName() + "(");

				ArrayList<String> args = new ArrayList<String>();
				ArrayList<String> rawargs = new ArrayList<String>();
				for(int i=0;i<p.arity();i++){
					args.add(id2str.get(rs.getInt(p.getArgs().get(i))));
					rawargs.add(rs.getString(p.getArgs().get(i)));
				}
				bw.append(StringMan.commaList(args) + ")</NAME>\n");
				bw.append("  <TYPE>" + "LR" + "</TYPE>\n");
				bw.append("  <LABEL>" + label + "</LABEL>\n");
				bw.append("  <PROB>" + probability + "</PROB>\n");
				bw.append("  <CORRECT>" + found + "</CORRECT>\n");

				
				// TODO: make this general
				// Get mention text
				String sql1 = "SELECT t0.word4 FROM pred_mention t0 WHERE t0.mid1=" + id1;
				System.out.println("mention query: " + sql1);
				ResultSet mid = db.query(sql1);
				String mention = "Missing";
				if(mid.next()) {
					mention = id2str.get(mid.getInt(1));
				}
				
				// Get entity text
				String sql2 = "SELECT t0.word2 FROM pred_entity t0 WHERE t0.eid1=" + id2;
				System.out.println("entity query: " + sql2);
				ResultSet eid = db.query(sql2);
				String entity = "NIL";
				if(eid.next()) {
					entity = id2str.get(eid.getInt(1));
				}
				
				// Get sentence text
				String sql3 = "SELECT t0.sentencetext2 FROM pred_sentence t0 WHERE t0.mid1=" + id1;
				System.out.println("sentence query: " + sql3);
				ResultSet sent = db.query(sql3);
				String sentence = "";
				if(sent.next()) {
					sentence = id2str.get(sent.getInt(1));
					sentence = sentence.replaceAll("&", "&amp;");
					sentence = sentence.replaceAll("<", "&lt;");
					sentence = sentence.replaceAll(">", "&gt;");
				}
				bw.append("  <COMMENT>\n");
				bw.append("    <DESP>Mention</DESP>\n");
				bw.append("    <RS><METARS>" + mention + "</METARS></RS>\n");
				bw.append("  </COMMENT>\n");
				bw.append("  <COMMENT>\n");
				bw.append("    <DESP>Entity</DESP>\n");
				bw.append("    <RS><METARS>" + entity + "</METARS></RS>\n");
				bw.append("  </COMMENT>\n");
				bw.append("  <COMMENT>\n");
				bw.append("    <DESP>Sentence</DESP>\n");
				bw.append("    <RS><METARS>" + sentence + "</METARS></RS>\n");
				bw.append("  </COMMENT>\n");

				
//				for(int i=1;i<=2;i++){
//					bw.append("  <COMMENT>\n");
//					bw.append("    <DESP>" + viewDesps[i-1] + "</DESP>\n");
//					String viewname = "_active_view" + i;
//					//ResultSet viewrs = db.query("SELECT * FROM " + viewname + " t0 WHERE " +
//					//		StringMan.join(" AND ", binds));
//					//while(viewrs.next()){
//					//	Integer[] columns = (Integer[])viewrs.getArray("viewcontent").getArray();
//					//	bw.append("    <RS>\n");
//					//	for(Integer ec : columns){
//					//		bw.append("      <METARS>" + id2str.get(ec) + "</METARS>\n");
//					//	}
//					//	bw.append("    </RS>\n");
//					bw.append("<RS><METARS></METARS></RS>\n");
//					//}
//					bw.append("  </COMMENT>\n");
//				}

				//bw.append("  <COMMENT>" + StringMan.commaList(rawargs) + "</COMMENT>\n");

				// Go through each rule
				while(provrs.next()){

					// This contains entity id
					smalllabelfields = new ArrayList<String>();
					for(int i=0;i<p.arity();i++){
						if(labelfields.contains(p.getArgs().get(i))){
							smalllabelfields.add(id2str.get(provrs.getInt(p.getArgs().get(i))));
						}
					}
					String thislabel = StringMan.commaList(smalllabelfields);

					int clauseID = provrs.getInt("prov");
					double weight = provrs.getFloat("weight");
					
					
					
					// NOTE: deepprov doesn't match up with terms
					
					String deepProv = provrs.getString("deepprov");	//{qid:160,docID:264066,queryWord:155}
					deepProv = deepProv.replaceAll("\\{|\\}", "");

					HashMap<String, Integer> term2constant = new HashMap<String, Integer>();
					String[] smalls = deepProv.split(",");
					//int num = 1;
					for(String small : smalls){
						String[] metamap = small.split(":");
						//String temp = metamap[0] + num;
						//term2constant.put(temp, Integer.valueOf(metamap[1]));
						//num++;
						term2constant.put(metamap[0], Integer.valueOf(metamap[1]));
					}

					// TODO: Be carefully with the consistency of clause id between different runs
 					FelixClause clause = fq.id2clause.get(clauseID);
					
					// Get source clause/comment
					String source_clause = clause.toString();
					String name = "";
					String content = "";
					
					String[] lines = source_clause.split("\n");
					for(String line : lines) {
						if(!line.equals("")) {
							if(line.contains(">")) {
								line = line.replaceAll(">", "&gt;");
							}
							if(line.contains("<")) {
								line = line.replaceAll("<", "&lt;");
							}
							if(line.startsWith("//")) {
								line = line.replaceFirst("//", "");
								line = line.trim();
								if(name.length() != 0) {
									name += "\n";
								}
								name += line;
							} else {
								if(content.length() != 0) {
									content += "\n";
								}
								content += line;
							}
						}
					}
				
					bw.append("  <PROV>\n");
					bw.append("  <SOURCECLAUSE>\n");
					bw.append("    <NAME>" + name + "</NAME>\n");
					bw.append("    <CONTENT>" + content + "</CONTENT>\n");
					bw.append("  </SOURCECLAUSE>\n");
					//bw.append("    <UNAME>" + clause.g)
					bw.append("    <WEIGHT>" + weight + "</WEIGHT>\n");
					bw.append("    <LABEL>" + thislabel + "</LABEL>\n");
					bw.append("    <CLAUSEID>" + clauseID + "</CLAUSEID>\n");
					
					// Go through each literal
					for(Literal l2 : clause.getRegLiterals()){
						Literal l = (Literal) l2.clone();

						// Rule for this predicate
						if(l.getPred().equals(p)){
							bw.append("    <LITERAL>\n");
							bw.append("      <SENSE>" + (l.getSense()==true ? "POS" : "NEG") + "</SENSE>\n");
							bw.append("      <NAME>" + l.getPred().getName() + "</NAME>\n");
							bw.append("      <STRINGFORM>=&gt;");
						}else{
							l.setSense(!l.getSense());			
							bw.append("    <LITERAL>\n");
							bw.append("      <SENSE>" + (l.getSense()==true ? "POS" : "NEG") + "</SENSE>\n");
							bw.append("      <NAME>" + l.getPred().getName() + "</NAME>\n");
							bw.append("      <STRINGFORM>");
						}

						// output predicate name
						bw.append(l.getSense()==true? "" : "!");
						bw.append(l.getPred().getName() + "(");
						
						ArrayList<String> terms = new ArrayList<String>();
						ArrayList<Integer> arguments = new ArrayList<Integer>();
						
						for(Term t : l.getTerms()){
							if(t.isConstant()){
								terms.add("\"" + id2str.get(Integer.valueOf(t.toString())) + "\"");
							}else{
								terms.add("\"" + id2str.get(term2constant.get(t.toString())) + "\"");
								arguments.add(term2constant.get(t.toString()));
							}
						}
						bw.append(StringMan.join(", ", terms) + ")</STRINGFORM>\n");

					
						// Go through each argument
						for(Integer id : arguments) {
							String str = "" + id2str.get(id);
							bw.append("      <ARG><NAME>" + str + "</NAME>");
							
							// TODO: make this general
							if(str.startsWith("MEN_")) {
								String sql_men = "SELECT t0.word4 FROM pred_mention t0 WHERE t0.mid1=" + id;
								ResultSet men = db.query(sql_men);
								if(men.next()) {
									String temp = id2str.get(men.getInt(1));
									bw.append("<VIEW><METARS>" + temp + "</METARS>");
								}
								String sql_sent = "SELECT t0.sentencetext2 FROM pred_sentence t0 WHERE t0.mid1=" + id;
								ResultSet sen = db.query(sql_sent);
								if(sen.next()) {
									String temp = id2str.get(sen.getInt(1));
									bw.append("<METARS>" + temp + "</METARS>");
								}
								bw.append("</VIEW>");
							} else if(str.startsWith("E_m_")) {
								String sql_ent = "SELECT t0.word2 FROM pred_entity t0 WHERE t0.eid1=" + id;
								ResultSet ent = db.query(sql_ent);
								if(ent.next()) {
									String temp = id2str.get(ent.getInt(1));
									temp = temp.replaceAll("&", "&amp;");
									temp = temp.replaceAll("<", "&lt;");
									temp = temp.replaceAll(">", "&gt;");
									bw.append("<VIEW><METARS>" + temp + "</METARS></VIEW>");
								}
							}
							bw.append("</ARG>\n");
						}
						bw.append("    </LITERAL>\n");
					}
					bw.append("  </PROV>\n");

					//System.out.print("\t" + provrs.getString("prov") + "\t" + provrs.getString("weight") + "\t" + provrs.getString("deepprov")+"\n\n");

				}

				bw.append("</QUERY>\n");
			}
		}
		bw.append("\n\n</root>\n\n");
		bw.close();

	}
	
	// TODO: make this general
	/**
	 * Creates tables/views in public schema for use by django feedback tool.
	 * @throws Exception
	 */
	public void DjangoSetup() throws Exception {
		
		HashMap<FelixPredicate, HashMap<String, String>> attrs = new 
				HashMap<FelixPredicate, HashMap<String, String>>();
		FelixQuery fq = new FelixQuery();
		ResultSet rs;
		String query = "";
		
		// For generating django models.py, admin,py
		ArrayList<String> models = new ArrayList<String>();
		ArrayList<String> admin = new ArrayList<String>();
		
		// setup stuff
		db = RDB.getRDBbyConfig(opt.evidDBSchema);
		db.resetSchema(opt.evidDBSchema);
		//db.resetSchema(FelixConfig.db_schema);
		String[] progFiles = opt.fprog.split(",");
		fq.loadPrograms(progFiles);
		

		HashMap<Long, String> id2str = db.loadIdSymbolMapFromTable();
		HashMap<String, Integer> str2id = db.loadSymbolIdMapFromTable();

		if(opt.fquery != null){
			String[] queryFiles = opt.fquery.split(",");
			fq.loadQueries(queryFiles);
		}
		
		if(opt.queryAtoms != null){
			FelixUIMan.println(">>> Parsing query atoms in command line");
			fq.parseQueryCommaList(opt.queryAtoms);
		}
		
		HashMap<Integer, Integer> gt = new HashMap<Integer, Integer>();

		// TODO: get this somewhere
		String predicate_name = "el";

		// gets metainfo
		rs = db.query("SELECT * FROM _prov_metainfo;");
		while(rs.next()){
			String pred = rs.getString(1);
			String attr = rs.getString(2);
			String value = rs.getString(3);

			FelixPredicate p = fq.getPredByName(pred);

			if(!attrs.containsKey(p)){
				attrs.put(p, new HashMap<String, String>());
			}
			attrs.get(p).put(attr, value);
		}
		
		FelixPredicate p = fq.getPredByName(predicate_name);
		ArrayList<String> arg_names = p.getArgs();

		// TODO: allow multiple predicates
		//for(FelixPredicate p : fq.getAllPred()){
		if(p.hasQuery()) {

			if(!attrs.get(p).get("optype").equals("lr")){
				ExceptionMan.die("Currently, I can only deal with LR!");
			}
			
			// this contains: eid2
			ArrayList<String> labelfields = new ArrayList<String>(Arrays.asList(attrs.get(p).get("labelfield").split(",")));
			
			
			// TODO: allow gt with different fields/key constraints
			// parse gt file
			String gtTable = "public." + p.getRelName() + "_gt";
			String gtTableTmp = "public." + p.getRelName() + "_gt_tmp";
			
			if(opt.gt != null) {
				// TODO: make this general
				// create tmp table for gt predicate
				db.dropTable(gtTableTmp);
				query = "CREATE TABLE " + gtTableTmp + " (mid1 integer, eid2 integer, mid1_value text, eid2_value text)";
				System.out.println(query);
				db.execute(query);
				
				
				System.out.println("Parsing gt file " + opt.gt);
				try {
					BufferedReader in = new BufferedReader(new FileReader(opt.gt));
					String str;
					while ((str = in.readLine()) != null) {
						if(predicate_name.equals(str.substring(0, str.indexOf('(')))) {
							
							String rest = str.substring(str.indexOf('(')+1, str.lastIndexOf(')') - str.indexOf('(') + 1);
							String[] args = rest.split(",");
	
							// Remove leading/trailing quotes
							for(int i = 0; i < args.length; i++) {
								args[i] = args[i].trim();
								if(args[i].startsWith("\"")) {
									args[i] = args[i].substring(1);
								}
								if(args[i].endsWith("\"")) {
									args[i] = args[i].substring(0, args[i].length() - 1);
								}
							}
							//System.out.println(args[0] + "\t" + args[1]);
							//System.out.println("\t" + str2id.get(args[0]) + "\t" + str2id.get(args[1]));
							
							// add to gt table
							Integer id1 = str2id.get(args[0]);
							Integer id2 = str2id.get(args[1]);
							
							query = "INSERT INTO " + gtTableTmp + " VALUES (" + id1 + ", " + id2 + ", '" + args[0] + "', 'n/" + args[1] + "')";
							System.out.println(query);
							db.execute(query);
						}
					}
					in.close();
					
					// TODO: make this general
					// create table for gt predicate
					db.dropTable(gtTable);
					query = "CREATE TABLE " + gtTable + " AS " +
							"SELECT t1.id AS id, t0.mid1 AS mid1, t0.eid2 AS eid2, t0.mid1_value AS mid1_value, t0.eid2_value AS eid2_value " +
							"FROM " + gtTableTmp + " AS t0, " + p.getRelName() + " AS t1 " +
							"WHERE t0.mid1 = t1.mid1 AND t0.eid2 = t1.eid2";
					System.out.println(query);
					db.execute(query);
					
					// drop tmp gt table
					query = "DROP TABLE " + gtTableTmp;
					System.out.println(query);
					db.execute(query);
				} catch (IOException e) {
					ExceptionMan.die("Error reading from file " + opt.gt);
				}
			}
			
			// TODO: make this general
			// Creates public view for this predicate
			db.dropView("public." + p.getRelName());
			if(gt != null) {
				query = "CREATE VIEW public." + p.getRelName() + " AS SELECT t0.*, t1.constantvalue as mid1_value, t2.constantvalue as eid2_value FROM " + p.getRelName() + " AS t0, type_mid AS t1, type_eid AS t2 WHERE t0.mid1 = t1.constantid AND t0.eid2 = t2.constantid AND mid1 IN (SELECT mid1 FROM " + gtTable + ")";
				System.out.println(query);
				db.execute(query);
				query = "ALTER TABLE " + gtTable + " ADD FOREIGN KEY (id) REFERENCES " + p.getRelName() + "(id)";
				System.out.println(query);
				db.execute(query);
			} else {
				query = "CREATE VIEW public." + p.getRelName() + " AS SELECT * FROM " + p.getRelName();
				System.out.println(query);
				db.execute(query);
			}
			
		
			String prov_table = attrs.get(p).get("provtable");
			
			if(prov_table == null) {
				ExceptionMan.die("Unable to find provtable for query predicate");
			}

			// TODO: make this general
			if(prov_table != null) {
				db.dropTable("public." + prov_table);
				if(opt.gt != null) {
					// TODO: make this step faster! (takes about 10 mins) 
					// maybe have command line option specifying if rules should be updated?
					query = "CREATE TABLE public." + prov_table + " AS SELECT t0.mid1, t0.eid2, t0.weight, t0.prov, array_to_string(t0.deepprov, ',') as deepprov, t1.id AS lid FROM " + prov_table + " AS t0, public." + p.getRelName() + " AS t1 WHERE t0.mid1 = t1.mid1 and t0.eid2 = t1.eid2";
					System.out.println(query);
					db.execute(query);
					query = "ALTER TABLE public." + prov_table + " ADD FOREIGN KEY (lid) REFERENCES " + p.getRelName() + "(id)";
					System.out.println(query);
					db.execute(query);
					query = "ALTER TABLE public." + prov_table + " ADD id SERIAL";
					System.out.println(query);
					db.update(query);
				} else {
					System.out.println("CREATE TABLE public." + prov_table + " AS SELECT * FROM " + prov_table);
					db.copyTable(prov_table, "public." + prov_table);
					query = "ALTER TABLE public." + prov_table + " ADD id SERIAL";
					System.out.println(query);
					db.update(query);
				}
			}
		}
		
		// Constants
		db.dropView("public.constants");
		query = "CREATE VIEW public.constants AS SELECT * FROM constants";
		System.out.println(query);
		db.execute(query);

		// Clauses
		db.dropTable("public.clauses");
		query = "CREATE TABLE public.clauses(id serial, comment text, content text)";
		System.out.println(query);
		db.execute(query);
		for(Integer id : fq.id2clause.keySet()) {
			FelixClause clause = fq.id2clause.get(id);
		
			int clauseID = clause.getId();
			// Get source clause/comment
			String source_clause = clause.toString();
			String comment = "";
			String content = "";
			
			String[] lines = source_clause.split("\n");
			for(String line : lines) {
				if(!line.equals("")) {
					line = line.replaceAll("'", "");
					if(line.startsWith("//")) {
						line = line.replaceFirst("//", "");
						line = line.trim();
						if(comment.length() != 0) {
							comment += "\n";
						}
						comment += line;
					} else {
						if(content.length() != 0) {
							content += "\n";
						}
						content += line;
					}
				}
			}
			
			query = "INSERT INTO public.clauses VALUES (" + clauseID + ", '" + comment + "', '" + content + "')";
			System.out.println(query);
			db.update(query);
		}
		
//		// TODO: go through all evidence files, create views (join with type tables)
//		for (Predicate pred : fq.getAllPred()) {
//			if(pred.hasEvid) {
//				String pred_name = pred.getRelName();
//				System.out.println("pred: " + pred_name); 
//			}
//			
//		}

		
		// Entity
		db.dropView("public.pred_entity");
		query = "CREATE VIEW public.pred_entity AS " +
				"SELECT t0.*, t1.constantvalue AS eid1_value, t2.constantvalue AS word2_value " +
				"FROM pred_entity AS t0, type_eid AS t1, type_word AS t2 " +
				"WHERE t0.eid1 = t1.constantid AND t0.word2 = t2.constantid";
		System.out.println(query);
		db.execute(query);
		
		// Mention
		db.dropView("public.pred_mention");
		query = "CREATE VIEW public.pred_mention AS " +
				"SELECT t0.*, t1.constantvalue AS mid1_value, t2.constantvalue AS sentid2_value, t3.constantvalue AS docid3_value, t4.constantvalue AS word4_value " +
				"FROM pred_mention AS t0, type_mid AS t1, type_sentid AS t2, type_docid AS t3, type_word AS t4 " +
				"WHERE t0.mid1 = t1.constantid AND t0.sentid2 = t2.constantid AND t0.docid3 = t3.constantid AND t0.word4 = t4.constantid";
		System.out.println(query);
		db.execute(query);
		
		// Sentence
		db.dropView("public.pred_sentence");
		query = "CREATE VIEW public.pred_sentence AS " +
				"SELECT t0.*, t1.constantvalue AS mid1_value, t2.constantvalue AS sentencetext2_value " +
				"FROM pred_sentence AS t0, type_mid AS t1, type_sentencetext AS t2 " +
				"WHERE t0.mid1 = t1.constantid AND t0.sentencetext2 = t2.constantid";
		System.out.println(query);
		db.execute(query);

		// TODO: auto-generate models.py
		
		// TODO: auto-generate admin.py
	}
	
	/**
	 * The constructor.
	 * @param _opt
	 */
	public ActiveLearner(FelixCommandOptions _opt){
		opt = _opt;
	}


}
