package felix.util;

import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.math.BigInteger;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.sql.DatabaseMetaData;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;

import org.postgresql.PGConnection;

import tuffy.db.RDB;
import tuffy.mln.Literal;
import tuffy.mln.Predicate;
import tuffy.mln.Term;
import tuffy.mln.Type;
import tuffy.ra.ConjunctiveQuery;
import tuffy.ra.Expression;
import tuffy.util.ExceptionMan;
import tuffy.util.StringMan;
import tuffy.util.UIMan;
import felix.dstruct.FelixClause;
import felix.dstruct.FelixPredicate;
import felix.dstruct.FelixQuery;
import felix.main.Felix;
import felix.parser.FelixCommandOptions;


public class CostCalculator {

	
	
	FelixCommandOptions opt;
	
	RDB db;
	
	/**
	 * Returns string containing given number of ones.
	 * @param _num
	 * @return
	 */
	public String ones(int _num){
		String rs = "";
		for(int i=0;i<_num;i++){
			rs = rs + "1";
		}
		return rs;
	}
	
	/**
	 * Returns string containing given number of zeros.
	 * @param _num
	 * @return
	 */
	public String zeros(int _num){
		String rs = "";
		for(int i=0;i<_num;i++){
			rs = rs + "0";
		}
		return rs;
	}
	
	/**
	 * Calculates cost.
	 * @throws Exception
	 */
	public void calc() throws Exception{

		Felix fx = new Felix();
		fx.options = opt;
		FelixQuery fq = fx.parseFelixQuery();
		fx.fq = fq;
		fx.loadEvidence();
		
		for(FelixPredicate p : fq.getPredicates()){
			if(p.getName().endsWith("_map")){
				//TODO: IF THERE IS P_MAP, ASSUME THERE ARE NO EVIDENCE FOR P
				String newName = p.getName().replaceAll("_map$", "");
				String sql = "DROP TABLE pred_" + newName;
				
				String typeName = p.getTypeAt(0).name();
				
				String newViewName = "pred_" + newName;
				
				Felix.db.execute(sql);
				
				sql = "CREATE VIEW " + newViewName + " AS SELECT 1 as id, TRUE as truth, 0 as prior, 2 as club, 0 as atomid, "
					 + " t0." + typeName + "1 AS " + typeName + "1, t1." + typeName + "1 AS " + typeName + "2 FROM " +
					 p.getRelName() + " t0, " + p.getRelName() + " t1 WHERE " + "t0." + typeName + "2 = t1." + typeName + "2";
				Felix.db.execute(sql);
			}
		}
		
		
		int ct = 0;
		int nHardBreak = 0;
		double totalCost = 0;
		for(FelixClause fc : fq.getAllClause()){
			
			//TODO: HARD CLAUSE
			if(fc.isHardClause()){
				continue;
			}
			

			
			HashSet<String> headTerms = new HashSet<String>();
			ConjunctiveQuery cq = new ConjunctiveQuery();
			HashMap<String, Type> var2type = new HashMap<String, Type>();
			
			for(Literal l : fc.getRegLiterals()){
				Literal l2 = (Literal) l.clone();
				l2.flipSense();
				cq.addBodyLit(l2);
				
				for(int i=0;i<l.getTerms().size();i++){
					Term t = l.getTerms().get(i);
					if(t.isVariable()){
						headTerms.add(t.var());
						var2type.put(t.var(), l.getPred().getTypeAt(i));
					}
				}
			}
			
			for(Expression e : fc.getConstraints()){
				cq.addConstraint(Expression.not(e));
			}
			
			FelixPredicate hp = new FelixPredicate("tmp_" + (ct++), true);
			for(String t : headTerms){
				hp.appendArgument(var2type.get(t));
			}
			hp.prepareDB(Felix.db);
			
			Literal l = new Literal(hp, true);
			for(String t : headTerms){
				l.appendTerm(new Term(t));
			}
			
			cq.setHead(l);
			cq.sourceClause = fc;
			cq.setWeight(fc.getWeight());
			
			System.err.println(fc);
			
			cq.buildIndexes(Felix.db, true, null, null, false, new ArrayList<String>(Arrays.asList("weight")), false);
			PreparedStatement ps = cq.psMap.get(this.zeros(l.getTerms().size()));
			
			System.out.println(ps);
			
			String sql = "CREATE TABLE " + "_for_cost_" + hp.getName() + " AS " + ps.toString();
			Felix.db.execute(sql);
			
			sql = "SELECT sum(weight) FROM " + "_for_cost_" + hp.getName();
			ResultSet rs = Felix.db.query(sql);

			double cost = 0;
			while(rs.next()){
				cost = rs.getDouble(1);
				break;
			}
			
			System.out.println();
			
			System.out.println("COST: [" + cost + "] for rule " + fc);
			
			totalCost += cost;
			
		}
		
		System.out.println("Total Cost = " + totalCost);
		

	}
	
	/**
	 * The constructor.
	 * @param _opt
	 */
	public CostCalculator(FelixCommandOptions _opt){
		opt = _opt;
	}

}
