package felix.util;

import org.kohsuke.args4j.CmdLineException;
import org.kohsuke.args4j.CmdLineParser;

import tuffy.parse.CommandOptions;
import tuffy.util.Config;
import tuffy.util.UIMan;
import felix.parser.FelixCommandOptions;

/**
 * Container of user-interface utilities.
 */
public class FelixUIMan extends UIMan{
	
	/**
	 * Returns string representation of given array.
	 * @param _array
	 * @return
	 */
    public static String joinArray(int[] _array){
    	if(_array.length == 0){
    		return "";
    	}
    	
    	String ret = "";
    	for(int i=0;i<_array.length - 1;i++){
    		ret = ret + _array[i] + ", ";
    	}
    	ret = ret + _array[_array.length-1];
    	return ret;
    }
	
    /**
     * Println with verbose level, indent.
     * @param verbose
     * @param indent
     * @param strings
     */
	public static void println(int verbose, int indent, String... strings){
		if(FelixConfig.verbose_level >= verbose){
					
			FelixConfig.currentIndent = indent;
			if(silent) return;
			
			
			if(Config.console_line_header != null){
				System.out.print("@" + Config.console_line_header + " ");
			}
			
			for(String s : strings){
				System.out.print(s);
				writeToDribbleFile(s);
			}
			System.out.println();
			writeToDribbleFile("\n");
			FelixConfig.currentIndent = 0;
		}
	}
	
	/**
	 * Println with verbose level.
	 * @param verbose
	 * @param strings
	 */
	public static void println(int verbose, String... strings){
		if(FelixConfig.verbose_level >= verbose){
			if(silent) return;
			
			if(Config.console_line_header != null){
				System.out.print("@" + Config.console_line_header + " ");
			}
			
			for(String s : strings){
				System.out.print(s);
				writeToDribbleFile(s);
			}
			System.out.println();
			writeToDribbleFile("\n");
		}
	}
	
	/**
	 * Print object with verbose level, indent.
	 * @param verbose
	 * @param indent
	 * @param obj
	 */
	public static <A> void printobj(int verbose, int indent, A obj){
		if(FelixConfig.verbose_level >= verbose){
			FelixConfig.currentIndent = indent;
			if(silent) return;
			FelixUIMan.println(verbose, indent, obj.toString());
			System.out.println();
			writeToDribbleFile("\n");
			FelixConfig.currentIndent = 0;
		}
	}

	/**
	 * Processes command line options.
	 * @param opt
	 * @return
	 */
	public static FelixCommandOptions processOptions(FelixCommandOptions opt){
		
		CommandOptions topt = UIMan.processOptions(opt);
		if(topt == null) return null;
		opt = (FelixCommandOptions) topt;
		FelixConfig.explainMode = opt.explainMode;
		//TODO set up more options
		return opt;
		
	}
	
	/**
	 * Parses command line options
	 * @param args
	 * @return FelixCommandOptions
	 */
	public static FelixCommandOptions parseCommand(String[] args){
		FelixCommandOptions opt = new FelixCommandOptions();
		CmdLineParser parser = new CmdLineParser(opt);
		try{
			parser.parseArgument(args);
			if(opt.showHelp){
				UIMan.println("USAGE:");
	            parser.printUsage(System.out);
	            return null;
			}
		}catch(CmdLineException e){
			System.err.println(e.getMessage());
			UIMan.println("USAGE:");
            parser.printUsage(System.out);
            return null;
		}

		opt = processOptions(opt);
		return opt;
	}

}
