package tuffy.mln;

/**
 * A term in first-order logic; either a variable or a constant.
 */
public class Term {
	
	/**
	 * Whether this term is a variable.
	 */
	private boolean isVariable;
	
	public boolean isOriAConstant = false;
	
	public int oriConstantID = -1;
	
	/**
	 * The name of this term. $var$ is not null iff.
	 * this term is a variable.
	 */
	private String var = null;

	/**
	 * The ID of this term. $constantID$ is not null iff.
	 * this term is a constant.
	 */
	private Integer constantID = null;
	
	
	/**
	 * Constructor of Term (variable version).
	 * 
	 * @param var name of the variable
	 */
	public Term(String var){
		isVariable = true;
		this.var = var;
	}
	
	
	/**
	 * Constructor a Term (constant version).
	 * 
	 * @param cid the constant in the form of its integer ID
	 */
	public Term(Integer cid){
		isVariable = false;
		constantID = cid;
	}
	
	/**
	 * Return whether this term is a variable.
	 */
	public boolean isVariable(){
		return isVariable;
	}
	
	/**
	 * 
	 * @return Whether this term is a constant.
	 */
	public boolean isConstant(){
		return !isVariable;
	}
	
	/**
	 * @return The name of this term, which is null when it is
	 * a constant.
	 */
	public String var(){
		return var;
	}
	
	/**
	 * 
	 * @return The name of this term, which is null when it is 
	 * a variable.
	 */
	public int constant(){
		return constantID;
	}
	
	/**
	 * @return This term's human-friendly representation.
	 */
	public String toString() {
		if(isVariable) {
			return var;
		}else {
			return Integer.toString(constantID);
		}
	}
	
}
